# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# Copyright (c) 2015, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#pylint: disable=R0904,C0103

import unittest
import httplib
import routes

import webob

from ztpserver.wsgiapp import WSGIController, WSGIRouter

class TestWsgiApp(unittest.TestCase):

    def setUp(self):
        mapper = routes.Mapper()
        mapper.collection('tests', 'test', controller=WSGIController)

        self.router = WSGIRouter(mapper)

    def request(self, url, method='GET', **kwargs):
        req = webob.Request.blank(url, method=method, **kwargs)
        return req.get_response(self.router)

    def get_url(self, url, expected_status=httplib.OK, **kwargs):
        resp = self.request(url, **kwargs)
        self.assertEqual(resp.status_code, expected_status)
        return resp

    def post_url(self, url, expected_status=httplib.OK, **kwargs):
        resp = self.request(url, 'POST', **kwargs)
        self.assertEqual(resp.status_code, expected_status)
        return resp

    def put_url(self, url, expected_status=httplib.OK, **kwargs):
        resp = self.request(url, 'PUT', **kwargs)
        self.assertEqual(resp.status_code, expected_status)
        return resp

    def delete_url(self, url, expected_status=httplib.OK, **kwargs):
        resp = self.request(url, 'DELETE', **kwargs)
        self.assertEqual(resp.status_code, expected_status)
        return resp

    def test_get_url_collection(self):
        self.get_url('/tests', 204)

    def test_get_url_resource(self):
        self.get_url('/test/resource', 404)

    def test_get_url_missing(self):
        self.get_url('/missing', 404)

    def test_post_url_collection(self):
        self.post_url('/tests', 204)

    def test_post_url_resource(self):
        self.post_url('/test/resource', 404)

    def test_post_url_missing(self):
        self.post_url('/missing', 404)

    def test_put_url_collection(self):
        self.put_url('/tests', 404)

    def test_put_url_resource(self):
        self.put_url('/test/resource', 404)

    def test_put_url_missing(self):
        self.put_url('/missing', 404)

    def test_delete_url_collection(self):
        self.delete_url('/tests', 404)

    def test_delete_url_resource(self):
        self.delete_url('/test/resource', 404)

    def test_delete_url_missing(self):
        self.delete_url('/missing', 404)

if __name__ == '__main__':
    unittest.main()






#!/usr/bin/env python
#
# Copyright (c) 2015, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#  - Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#  - Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#  - Neither the name of Arista Networks nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#pylint: disable=R0904,F0401,W0232,E1101,W0402

import os
import os.path
import unittest
import sys
from string import Template

sys.path.append('test/client')

from client_test_lib import Bootstrap, ActionFailureTest
from client_test_lib import file_log, get_action, random_string
from client_test_lib import startup_config_action
from client_test_lib import raise_exception

class FailureTest(ActionFailureTest):

    def test_missing_url(self):
        self.basic_test('add_config', 'Missing attribute(\'url\')')

    def test_url_failure(self):
        self.basic_test('add_config', 'Unable to retrieve config from URL',
                        attributes={'url' :
                                    random_string()})

    def test_variables_failure(self):
        url = random_string()
        contents = random_string()
        self.basic_test('add_config', 
                        'Unable to perform variable substitution - '
                        'invalid variables',
                        attributes={'url' : url,
                                    'variables' : random_string()},
                        file_responses={url : contents})

    def test_variable_missing_failure(self):
        url = random_string()
        contents = random_string() + ' $missing_var'
        self.basic_test('add_config', 
                        'Unable to perform variable substitution - '
                        '\'missing_var\' missing from list of substitutions',
                        attributes={'url' : url,
                                    'substitution_mode': 'strict',
                                    'variables' : {}},
                        file_responses={url : contents})

    def test_invalid_substitution_mode(self):
        self.basic_test('add_config', 
                        'Invalid option specified for '
                        'substitution_mode attribute',
                        attributes={'url': random_string(),
                                    'substitution_mode': 'dummy'})


class SuccessTest(unittest.TestCase):


    def test_success(self):
        bootstrap = Bootstrap(ztps_default_config=True)
        config = random_string()
        url = config
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'test_action',
                      'attributes': {'url' : url}}])
        bootstrap.ztps.set_action_response('test_action',
                                           get_action('add_config'))
        contents = random_string()
        bootstrap.ztps.set_file_response(config, contents)
        bootstrap.start_test()

        try:
            self.failUnless(os.path.isfile(bootstrap.startup_config))
            self.failUnless(contents.split() == 
                            file_log(bootstrap.startup_config))
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            bootstrap.end_test()

    def test_success_url(self):
        bootstrap = Bootstrap(ztps_default_config=True)
        config = random_string()
        url = 'http://%s/%s' % (bootstrap.server, config)
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'test_action',
                      'attributes': {'url' : url}}])
        bootstrap.ztps.set_action_response('test_action',
                                           get_action('add_config'))
        contents = random_string()
        bootstrap.ztps.set_file_response(config, contents)
        bootstrap.start_test()

        try:
            self.failUnless(os.path.isfile(bootstrap.startup_config))
            self.failUnless(contents.split() == 
                            file_log(bootstrap.startup_config))
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            bootstrap.end_test()

    def test_append(self):
        bootstrap = Bootstrap(ztps_default_config=True)
        config = random_string()
        url = 'http://%s/%s' % (bootstrap.server, config)
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'startup_config_action'},
                     {'action' : 'test_action',
                      'attributes': {'url' : url}}])
        bootstrap.ztps.set_action_response('test_action',
                                           get_action('add_config'))

        startup_config_text = random_string()
        bootstrap.ztps.set_action_response(
            'startup_config_action',
            startup_config_action(lines=[startup_config_text]))
        contents = random_string()
        bootstrap.ztps.set_file_response(config, contents)
        bootstrap.start_test()

        try:
            self.failUnless(os.path.isfile(bootstrap.startup_config))
            log = file_log(bootstrap.startup_config)
            self.failUnless(contents in log)
            self.failUnless(startup_config_text in log)
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            bootstrap.end_test()

    def test_multi_lines(self):
        bootstrap = Bootstrap(ztps_default_config=True)
        config = random_string()
        url = 'http://%s/%s' % (bootstrap.server, config)
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'startup_config_action'},
                     {'action' : 'test_action',
                      'attributes': {'url' : url}}])
        bootstrap.ztps.set_action_response('test_action',
                                           get_action('add_config'))

        startup_config_lines = [random_string(), random_string(),
                                random_string(), random_string()]
        bootstrap.ztps.set_action_response(
            'startup_config_action',
            startup_config_action(lines=startup_config_lines))
        contents = '\n'.join([random_string(), random_string(),
                              random_string(), random_string(),
                              random_string(), random_string()])
        bootstrap.ztps.set_file_response(config, contents)
        bootstrap.start_test()

        try:
            self.failUnless(os.path.isfile(bootstrap.startup_config))
            log = file_log(bootstrap.startup_config)
            all_lines = startup_config_lines + contents.split()
            for line in all_lines:
                self.failUnless(line in log)
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            bootstrap.end_test()

    def test_variables_strict(self):
        bootstrap = Bootstrap(ztps_default_config=True)
        config = random_string()
        url = config
        var_dict = { 'a' : 'A',
                     'b' : 'A',
                     'xxx' : '999',
                     'dummy': 'DUMMY'}
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'test_action',
                      'attributes': {'url' : url,
                                     'substitution_mode': 'strict',
                                     'variables': var_dict}}])
        bootstrap.ztps.set_action_response('test_action',
                                           get_action('add_config'))
        contents = '$a 1234 $b 4  321 $xxx$a'
        expected_contents = Template(contents).substitute(var_dict)
        bootstrap.ztps.set_file_response(config, contents)
        bootstrap.start_test()

        try:
            self.failUnless(os.path.isfile(bootstrap.startup_config))
            self.failUnless([expected_contents] ==
                            file_log(bootstrap.startup_config))
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            bootstrap.end_test()

    def test_variables_loose(self):
        bootstrap = Bootstrap(ztps_default_config=True)
        config = random_string()
        url = config
        var_dict = { 'a' : 'A',
                     'b' : 'A',
                     'xxx' : '999',
                     'dummy': 'DUMMY'}
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'test_action',
                      'attributes': {'url' : url,
                                     'substitution_mode': 'loose',
                                     'variables': var_dict}}])
        bootstrap.ztps.set_action_response('test_action',
                                           get_action('add_config'))
        contents = '$a 1234 $b 4  321 $xxx$a'
        expected_contents = Template(contents).safe_substitute(var_dict)
        bootstrap.ztps.set_file_response(config, contents)
        bootstrap.start_test()

        try:
            self.failUnless(os.path.isfile(bootstrap.startup_config))
            self.failUnless([expected_contents] ==
                            file_log(bootstrap.startup_config))
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            bootstrap.end_test()

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c) 2015, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#  - Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#  - Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#  - Neither the name of Arista Networks nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#pylint: disable=R0904,F0401,W0232,E1101

import os
import os.path
import random
import shutil
import sys
import unittest

from stat import ST_MODE

sys.path.append('test/client')

from client_test_lib import Bootstrap, ActionFailureTest
from client_test_lib import file_log, get_action, random_string
from client_test_lib import startup_config_action, remove_file
from client_test_lib import raise_exception

def random_permissions():
    return '7%s%s' % ((random.choice([1, 2, 3, 4, 5, 6, 7]),
                       random.choice([1, 2, 3, 4, 5, 6, 7])))


class FailureTest(ActionFailureTest):

    def test_missing_src_url(self):
        self.basic_test('copy_file',
                        'Missing attribute(\'src_url\')')

    def test_missing_dst_url(self):
        self.basic_test('copy_file',
                        'Missing attribute(\'dst_url\')',
                        attributes={'src_url' :
                                        random_string()})

    def test_wrong_overwrite_value(self):
        self.basic_test('copy_file',
                        'Erroneous \'overwrite\' value',
                        attributes={'src_url' :
                                    random_string(),
                                    'dst_url' :
                                    random_string(),
                                    'overwrite' :
                                    'bogus'})

    def test_url_failure(self):
        action = get_action('copy_file')
        action = action.replace('/mnt/flash/.ztp-files',
                                '/tmp')

        self.basic_test('copy_file',
                        'Unable to retrieve file from URL',
                        attributes={'src_url' :
                                    random_string(),
                                    'dst_url' :
                                    random_string()},
                        action_value=action)


class SuccessSrcUrlReplacementTests(unittest.TestCase):

    def test_success(self):
        bootstrap = Bootstrap(ztps_default_config=True)

        source = random_string()
        destination = '/tmp/%s' % random_string()
        ztps_server = 'http://%s' % bootstrap.server
        url = 'http://%s/%s' % (bootstrap.server, source)

        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'startup_config_action'},
                     {'action' : 'test_action',
                      'attributes' : {'src_url'    : url,
                                      'dst_url'    : destination,
                                      'ztps_server': ztps_server}}])

        bootstrap.ztps.set_action_response(
            'startup_config_action', startup_config_action())

        action = get_action('copy_file')

        # Make the destinaton persistent
        action = action.replace('PERSISTENT_STORAGE = [',
                                'PERSISTENT_STORAGE = [\'%s\', ' %
                                destination)
        bootstrap.ztps.set_action_response('test_action',
                                           action)

        contents = random_string()
        bootstrap.ztps.set_file_response(source, contents)

        bootstrap.start_test()

        destination_path = '%s/%s' % (destination, source)

        try:
            self.failUnless(os.path.isfile(destination_path))
            self.failUnless([contents] ==
                            file_log(destination_path))
            self.failIf(os.path.isfile(bootstrap.rc_eos))
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            remove_file(destination_path)
            shutil.rmtree(destination)
            bootstrap.end_test()

    def test_append_server(self):
        bootstrap = Bootstrap(ztps_default_config=True)

        source = random_string()
        destination = '/tmp/%s' % random_string()
        ztps_server = 'http://%s' % bootstrap.server

        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'startup_config_action'},
                     {'action' : 'test_action',
                     'attributes' : {'src_url'    : source,
                                     'dst_url'    : destination,
                                     'ztps_server': ztps_server}}])

        bootstrap.ztps.set_action_response(
            'startup_config_action', startup_config_action())

        action = get_action('copy_file')

        # Make the destinaton persistent
        action = action.replace('PERSISTENT_STORAGE = [',
                                'PERSISTENT_STORAGE = [\'%s\', ' %
                                destination)
        bootstrap.ztps.set_action_response('test_action',
                                           action)

        contents = random_string()
        bootstrap.ztps.set_file_response(source, contents)

        bootstrap.start_test()

        destination_path = '%s/%s' % (destination, source)

        try:
            self.failUnless(os.path.isfile(destination_path))
            self.failUnless([contents] ==
                            file_log(destination_path))
            self.failIf(os.path.isfile(bootstrap.rc_eos))
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            remove_file(destination_path)
            shutil.rmtree(destination)
            bootstrap.end_test()




class SuccessPersistentTest(unittest.TestCase):

    def test_success(self):
        bootstrap = Bootstrap(ztps_default_config=True)

        source = random_string()
        destination = '/tmp/%s' % random_string()

        url = 'http://%s/%s' % (bootstrap.server, source)
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'startup_config_action'},
                     {'action' : 'test_action',
                      'attributes' : {'src_url' : url,
                                      'dst_url' : destination}}])

        bootstrap.ztps.set_action_response(
            'startup_config_action', startup_config_action())

        action = get_action('copy_file')

        # Make the destinaton persistent
        action = action.replace('PERSISTENT_STORAGE = [',
                                'PERSISTENT_STORAGE = [\'%s\', ' %
                                destination)
        bootstrap.ztps.set_action_response('test_action',
                                           action)

        contents = random_string()
        bootstrap.ztps.set_file_response(source, contents)

        bootstrap.start_test()

        destination_path = '%s/%s' % (destination, source)

        try:
            self.failUnless(os.path.isfile(destination_path))
            self.failUnless([contents] ==
                            file_log(destination_path))
            self.failIf(os.path.isfile(bootstrap.rc_eos))
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            remove_file(destination_path)
            shutil.rmtree(destination)
            bootstrap.end_test()

    def test_replace(self):
        bootstrap = Bootstrap(ztps_default_config=True)

        source = random_string()
        destination = '/tmp/%s' % random_string()

        url = 'http://%s/%s' % (bootstrap.server, source)
        attributes = {'src_url' : url,
                      'dst_url' : destination}

        # 'replace' is the default
        if bool(random.getrandbits(1)):
            attributes['overwrite'] = 'replace'

        mode = None
        if True or bool(random.getrandbits(1)):
            mode = random_permissions()
            attributes['mode'] = mode

        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'startup_config_action'},
                     {'action' : 'test_action',
                      'attributes' : attributes}])

        bootstrap.ztps.set_action_response(
            'startup_config_action', startup_config_action())

        action = get_action('copy_file')

        # Make the destinaton persistent
        action = action.replace('PERSISTENT_STORAGE = [',
                                'PERSISTENT_STORAGE = [\'%s\', ' %
                                destination)
        bootstrap.ztps.set_action_response('test_action',
                                           action)

        contents = random_string()
        bootstrap.ztps.set_file_response(source, contents)

        destination_path = '%s/%s' % (destination, source)
        bootstrap.start_test()

        try:
            self.failUnless(os.path.isfile(destination_path))
            self.failUnless([contents] ==
                            file_log(destination_path))
            self.failIf(os.path.isfile(bootstrap.rc_eos))
            if mode:
                self.failUnless(mode ==
                                oct(os.stat(destination_path)[ST_MODE])[-3:])
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            remove_file(destination_path)
            shutil.rmtree(destination)
            bootstrap.end_test()

    def test_keep_original(self):
        bootstrap = Bootstrap(ztps_default_config=True)

        source = random_string()
        destination = '/tmp/%s' % random_string()

        url = 'http://%s/%s' % (bootstrap.server, source)
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'startup_config_action'},
                     {'action' : 'test_action',
                      'attributes' : {'src_url' : url,
                                      'dst_url' : destination,
                                      'overwrite' : 'if-missing'}}])

        bootstrap.ztps.set_action_response(
            'startup_config_action', startup_config_action())

        action = get_action('copy_file')

        # Make the destinaton persistent
        action = action.replace('PERSISTENT_STORAGE = [',
                                'PERSISTENT_STORAGE = [\'%s\', ' %
                                destination)
        bootstrap.ztps.set_action_response('test_action',
                                           action)

        contents = random_string()
        bootstrap.ztps.set_file_response(source, contents)

        destination_path = '%s/%s' % (destination, source)
        existing_contents = random_string()
        os.makedirs(destination)
        file_descriptor = open(destination_path, 'w')
        file_descriptor.write(existing_contents)
        file_descriptor.close()

        bootstrap.start_test()

        try:
            self.failUnless(os.path.isfile(destination_path))
            self.failUnless([existing_contents] ==
                            file_log(destination_path))

            self.failIf(os.path.isfile(bootstrap.rc_eos))
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            remove_file(destination_path)
            shutil.rmtree(destination)
            bootstrap.end_test()

    def test_keep_backup(self):
        bootstrap = Bootstrap(ztps_default_config=True)

        source = random_string()
        destination = '/tmp/%s' % random_string()

        url = 'http://%s/%s' % (bootstrap.server, source)
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'startup_config_action'},
                     {'action' : 'test_action',
                      'attributes' : {'src_url' : url,
                                      'dst_url' : destination,
                                      'overwrite' : 'backup'}}])

        bootstrap.ztps.set_action_response(
            'startup_config_action', startup_config_action())

        action = get_action('copy_file')

        # Make the destinaton persistent
        action = action.replace('PERSISTENT_STORAGE = [',
                                'PERSISTENT_STORAGE = [\'%s\', ' %
                                destination)
        bootstrap.ztps.set_action_response('test_action',
                                           action)

        contents = random_string()
        bootstrap.ztps.set_file_response(source, contents)

        destination_path = '%s/%s' % (destination, source)
        backup_contents = random_string()
        os.makedirs(destination)
        file_descriptor = open(destination_path, 'w')
        file_descriptor.write(backup_contents)
        file_descriptor.close()

        bootstrap.start_test()

        backup_path = '%s.backup' % destination_path
        try:
            self.failUnless(os.path.isfile(destination_path))
            self.failUnless([contents] ==
                            file_log(destination_path))

            self.failUnless(os.path.isfile(backup_path))
            self.failUnless([backup_contents] ==
                            file_log(backup_path))

            self.failIf(os.path.isfile(bootstrap.rc_eos))
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            remove_file(destination_path)
            remove_file(backup_path)
            shutil.rmtree(destination)
            bootstrap.end_test()

class SuccessNonPersistentTest(unittest.TestCase):

    def test_success(self):
        bootstrap = Bootstrap(ztps_default_config=True)

        source = random_string()
        destination = random_string()

        url = 'http://%s/%s' % (bootstrap.server, source)
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'startup_config_action'},
                     {'action' : 'test_action',
                      'attributes' : {'src_url' : url,
                                      'dst_url' : destination}}])

        bootstrap.ztps.set_action_response(
            'startup_config_action', startup_config_action())

        persistent_dir = '/tmp'
        action = get_action('copy_file')
        action = action.replace('/mnt/flash/.ztp-files',
                                persistent_dir)
        bootstrap.ztps.set_action_response('test_action',
                                           action)

        contents = random_string()
        bootstrap.ztps.set_file_response(source, contents)

        bootstrap.start_test()

        destination_path = '%s/%s' % (persistent_dir, source)

        try:
            self.failUnless(os.path.isfile(destination_path))
            self.failUnless([contents] ==
                            file_log(destination_path))

            self.failUnless(os.path.isfile(bootstrap.rc_eos))
            log = file_log(bootstrap.rc_eos)
            self.failUnless('#!/bin/bash' in log)
            self.failUnless('sudo cp %s %s' %
                            (destination_path, destination) in log)
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            remove_file(destination_path)
            bootstrap.end_test()

    def test_replace(self):
        bootstrap = Bootstrap(ztps_default_config=True)

        source = random_string()
        destination = random_string()

        url = 'http://%s/%s' % (bootstrap.server, source)
        attributes = {'src_url' : url,
                      'dst_url' : destination}

        # 'replace' is the default
        if bool(random.getrandbits(1)):
            attributes['overwrite'] = 'replace'

        mode = None
        if True or bool(random.getrandbits(1)):
            mode = random_permissions()
            attributes['mode'] = mode

        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'startup_config_action'},
                     {'action' : 'test_action',
                      'attributes' : attributes}])

        bootstrap.ztps.set_action_response(
            'startup_config_action', startup_config_action())

        persistent_dir = '/tmp'
        action = get_action('copy_file')
        action = action.replace('/mnt/flash/.ztp-files',
                                persistent_dir)
        bootstrap.ztps.set_action_response('test_action',
                                           action)

        contents = random_string()
        bootstrap.ztps.set_file_response(source, contents)

        destination_path = '%s/%s' % (persistent_dir, source)
        bootstrap.start_test()

        try:
            self.failUnless(os.path.isfile(destination_path))
            self.failUnless([contents] ==
                            file_log(destination_path))

            self.failUnless(os.path.isfile(bootstrap.rc_eos))
            log = file_log(bootstrap.rc_eos)
            self.failUnless('#!/bin/bash' in log)
            self.failUnless('sudo cp %s %s' %
                            (destination_path, destination) in log)
            if mode:
                self.failUnless('sudo chmod %s %s' %
                                (mode, destination) in log)
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            remove_file(destination_path)
            bootstrap.end_test()

    def test_keep_original(self):
        bootstrap = Bootstrap(ztps_default_config=True)

        source = random_string()
        destination = random_string()

        url = 'http://%s/%s' % (bootstrap.server, source)
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'startup_config_action'},
                     {'action' : 'test_action',
                      'attributes' : {'src_url' : url,
                                      'dst_url' : destination,
                                      'overwrite' : 'if-missing'}}])

        bootstrap.ztps.set_action_response(
            'startup_config_action', startup_config_action())

        persistent_dir = '/tmp'
        action = get_action('copy_file')
        action = action.replace('/mnt/flash/.ztp-files',
                                persistent_dir)
        bootstrap.ztps.set_action_response('test_action',
                                           action)

        contents = random_string()
        bootstrap.ztps.set_file_response(source, contents)

        bootstrap.start_test()

        destination_path = '%s/%s' % (persistent_dir, source)
        try:
            self.failUnless(os.path.isfile(destination_path))
            self.failUnless([contents] ==
                            file_log(destination_path))

            self.failUnless(os.path.isfile(bootstrap.rc_eos))
            log = file_log(bootstrap.rc_eos)
            self.failUnless('#!/bin/bash' in log)
            self.failUnless('[ ! -f %s ] && sudo cp %s %s' %
                            (destination, destination_path,
                             destination) in log)
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            remove_file(destination_path)
            bootstrap.end_test()

    def test_keep_backup(self):
        bootstrap = Bootstrap(ztps_default_config=True)

        source = random_string()
        destination = random_string()

        url = 'http://%s/%s' % (bootstrap.server, source)
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'startup_config_action'},
                     {'action' : 'test_action',
                      'attributes' : {'src_url' : url,
                                      'dst_url' : destination,
                                      'overwrite' : 'backup'}}])

        bootstrap.ztps.set_action_response(
            'startup_config_action', startup_config_action())

        persistent_dir = '/tmp'
        action = get_action('copy_file')
        action = action.replace('/mnt/flash/.ztp-files',
                                persistent_dir)
        bootstrap.ztps.set_action_response('test_action',
                                           action)

        contents = random_string()
        bootstrap.ztps.set_file_response(source, contents)

        bootstrap.start_test()

        destination_path = '%s/%s' % (persistent_dir, source)
        try:
            self.failUnless(os.path.isfile(destination_path))
            self.failUnless([contents] ==
                            file_log(destination_path))

            self.failUnless(os.path.isfile(bootstrap.rc_eos))
            log = file_log(bootstrap.rc_eos)
            self.failUnless('#!/bin/bash' in log)
            self.failUnless('sudo cp %s %s' %
                            (destination_path, destination) in log)
            self.failUnless('[ -f %s ] && sudo mv %s %s.backup' %
                            (destination, destination,
                             destination) in log)
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            remove_file(destination_path)
            bootstrap.end_test()

if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2015, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# pylint: disable=C0102,C0103,E1103,W0142,W0613,C0302,E1120
#

import json
import random
import unittest

from webob import Request

from mock import MagicMock, Mock, patch

import ztpserver.controller
import ztpserver.config
import ztpserver.repository

from ztpserver.controller import DEFINITION_FN, PATTERN_FN

from ztpserver.repository import FileObjectNotFound, FileObjectError

from server_test_lib import enable_logging, remove_all, random_string
from server_test_lib import mock_match, ztp_headers, write_file
from server_test_lib import create_definition, create_attributes, create_node
from server_test_lib import create_bootstrap_conf

import ztpserver.constants as constants


class RouterUnitTests(unittest.TestCase):

    def match_routes(self, url, valid, invalid):

        request = Request.blank(url)
        router = ztpserver.controller.Router()

        if valid:
            for method in valid.split(','):
                request.method = method
                msg = 'method %s failed for url %s' % (method, url)
                self.assertIsNotNone(router.map.match(environ=request.environ),
                                     msg)

        if invalid:
            for method in invalid.split(','):
                request.method = method
                msg = 'method %s failed for url %s' % (method, url)
                self.assertIsNone(router.map.match(environ=request.environ),
                                  msg)

    def test_bootstrap_collection(self):
        url = '/bootstrap'
        self.match_routes(url, 'GET', 'POST,PUT,DELETE')

    def test_bootstrap_resource(self):
        url = '/bootstrap/%s' % random_string()
        self.match_routes(url, None, 'GET,POST,PUT,DELETE')

    def test_bootstrap_config(self):
        url = '/bootstrap/config'
        self.match_routes(url, 'GET', 'POST,PUT,DELETE')

    def test_meta_action(self):
        url = '/meta/actions/dummy'
        self.match_routes(url, 'GET', 'POST,PUT,DELETE')

    def test_meta_files(self):
        url = '/meta/files/dummy'
        self.match_routes(url, 'GET', 'POST,PUT,DELETE')

    def test_meta_files_long(self):
        url = '/meta/files/dummy/long'
        self.match_routes(url, 'GET', 'POST,PUT,DELETE')

    def test_files_collection(self):
        url = '/files'
        self.match_routes(url, None, 'GET,POST,PUT,DELETE')

    def test_files_resource(self):
        url = '/files/%s' % random_string()
        self.match_routes(url, 'GET', 'POST,PUT,DELETE')

        url = '/files/%s/%s' % (random_string(), random_string())
        self.match_routes(url, 'GET', 'POST,PUT,DELETE')

    def test_actions_collection(self):
        url = '/actions'
        self.match_routes(url, None, 'GET,POST,PUT,DELETE')

    def test_actions_resource(self):
        url = '/actions/%s' % random_string()
        self.match_routes(url, 'GET', 'POST,PUT,DELETE')

    def test_nodes_collection(self):
        url = '/nodes'
        self.match_routes(url, 'POST', 'GET,PUT,DELETE')

    def test_nodes_resource(self):
        url = '/nodes/%s' % random_string()
        self.match_routes(url, 'GET', 'POST,PUT,DELETE')

    def test_nodes_resource_get_config(self):
        url = '/nodes/%s/startup-config' % random_string()
        self.match_routes(url, 'GET,PUT', 'POST,DELETE')



class MetaControllerUnitTests(unittest.TestCase):

    @patch('ztpserver.controller.create_repository')
    def test_bad_request_file_not_found(self, m_repository):
        error = IOError
        cfg = {'return_value.get_file.side_effect': error}
        m_repository.configure_mock(**cfg)

        path = '/'.join([random_string()] * random.randint(1, 4))

        controller = ztpserver.controller.MetaController()
        resp = controller.metadata(None,
                                   type=random.choice(['files', 'actions']),
                                   path_info=path)

        self.assertEqual(resp['body'], '')
        self.assertEqual(resp['content_type'], constants.CONTENT_TYPE_HTML)
        self.assertEqual(resp['status'], constants.HTTP_STATUS_NOT_FOUND)

    @patch('ztpserver.controller.create_repository')
    def test_bad_request_io_error(self, m_repository):
        cfg = random.choice([
                {'return_value.get_file.return_value.hash.side_effect':
                 IOError},
                {'return_value.get_file.return_value.size.side_effect':
                 IOError}])
        m_repository.configure_mock(**cfg)

        controller = ztpserver.controller.MetaController()
        resp = controller.metadata(None,
                                   type=random.choice(['files', 'actions']),
                                   path_info=random_string())

        self.assertEqual(resp['body'], '')
        self.assertEqual(resp['content_type'], constants.CONTENT_TYPE_HTML)
        self.assertEqual(resp['status'],
                         constants.HTTP_STATUS_INTERNAL_SERVER_ERROR)

    @patch('ztpserver.controller.create_repository')
    def test_success(self, m_repository):
        sha1 = random_string()
        size = random.randint(1, 1000000)
        cfg = {'return_value.get_file.return_value.hash.return_value':
               sha1,
               'return_value.get_file.return_value.size.return_value':
               size}
        m_repository.configure_mock(**cfg)

        controller = ztpserver.controller.MetaController()
        resp = controller.metadata(None,
                                   type=random.choice(['files', 'actions']),
                                   path_info=random_string())

        self.assertEqual(resp['body'], {'sha1': sha1, 'size': size})
        self.assertEqual(resp['content_type'], constants.CONTENT_TYPE_JSON)


class BootstrapConfigUnitTests(unittest.TestCase):

    @patch('ztpserver.controller.create_repository')
    @patch('string.Template.safe_substitute')
    def test_index_success(self, m_substitute, m_repository):
        m_substitute.return_value = random_string()

        controller = ztpserver.controller.BootstrapController()

        request = Request.blank('')
        request.remote_addr = ''
        resp = controller.index(request)

        self.assertTrue(m_substitute.called)
        self.assertEqual(resp['content_type'], constants.CONTENT_TYPE_PYTHON)
        self.assertEqual(resp['body'], m_substitute.return_value)

    @patch('ztpserver.controller.create_repository')
    def test_index_bootstrap_not_found_failure(self, m_repository):
        cfg = {'return_value.get_file.side_effect': FileObjectError}
        m_repository.configure_mock(**cfg)

        controller = ztpserver.controller.BootstrapController()
        resp = controller.index(None)

        self.assertTrue(resp['status'], constants.HTTP_STATUS_BAD_REQUEST)
        self.assertEqual(resp['body'], '')

    @patch('ztpserver.controller.create_repository')
    def test_index_bootstrap_inaccessible_failure(self, m_repository):
        cfg = {'return_value.read.side_effect': FileObjectError}
        m_repository.return_value.get_file.configure_mock(**cfg)

        controller = ztpserver.controller.BootstrapController()
        resp = controller.index(None)

        self.assertTrue(resp['status'], constants.HTTP_STATUS_BAD_REQUEST)
        self.assertEqual(resp['body'], '')

    @patch('ztpserver.controller.create_repository')
    def test_config_success(self, m_repository):
        config = create_bootstrap_conf()
        config.add_logging(dict(destination=random_string(),
                                level=random_string()))

        cfg = {'return_value.read.return_value': config.as_dict()}
        m_repository.return_value.get_file.configure_mock(**cfg)

        controller = ztpserver.controller.BootstrapController()

        request = Request.blank('')
        request.remote_addr = ''
        resp = controller.config(request)

        self.assertEqual(resp['body'], config.as_dict())
        self.assertEqual(resp['content_type'], constants.CONTENT_TYPE_JSON)

    @patch('ztpserver.controller.create_repository')
    def test_config_defaults(self, m_repository):
        cfg = {'return_value.get_file.side_effect': FileObjectNotFound}
        m_repository.configure_mock(**cfg)

        controller = ztpserver.controller.BootstrapController()

        request = Request.blank('')
        request.remote_addr = ''
        resp = controller.config(request)

        self.assertEqual(resp['body'], controller.DEFAULT_CONFIG)
        self.assertEqual(resp['content_type'], constants.CONTENT_TYPE_JSON)

    @patch('ztpserver.controller.create_repository')
    def test_config_failure(self, m_repository):
        cfg = {'return_value.read.side_effect': FileObjectError}
        m_repository.return_value.get_file.configure_mock(**cfg)

        controller = ztpserver.controller.BootstrapController()
        resp = controller.config(None)

        self.assertEqual(resp['body'], '')
        self.assertEqual(resp['status'], constants.HTTP_STATUS_BAD_REQUEST)

    @patch('ztpserver.controller.create_repository')
    def test_no_config(self, m_repository):
        cfg = {'return_value.read.return_value': {}}
        m_repository.return_value.get_file.configure_mock(**cfg)

        controller = ztpserver.controller.BootstrapController()

        request = Request.blank('')
        request.remote_addr = ''
        resp = controller.config(request)

        self.assertEqual(resp['body'], controller.DEFAULT_CONFIG)
        self.assertEqual(resp['content_type'], constants.CONTENT_TYPE_JSON)

    @patch('ztpserver.controller.create_repository')
    def test_no_xmpp(self, m_repository):
        cfg = {'return_value.read.return_value': {'logging': []}}
        m_repository.return_value.get_file.configure_mock(**cfg)

        controller = ztpserver.controller.BootstrapController()

        request = Request.blank('')
        request.remote_addr = ''
        resp = controller.config(request)

        self.assertEqual(resp['body'], controller.DEFAULT_CONFIG)
        self.assertEqual(resp['content_type'], constants.CONTENT_TYPE_JSON)

    @patch('ztpserver.controller.create_repository')
    def test_no_logging(self, m_repository):
        cfg = {'return_value.read.return_value': {'xmpp': {}}}
        m_repository.return_value.get_file.configure_mock(**cfg)

        controller = ztpserver.controller.BootstrapController()

        request = Request.blank('')
        request.remote_addr = ''
        resp = controller.config(request)

        self.assertEqual(resp['body'], controller.DEFAULT_CONFIG)
        self.assertEqual(resp['content_type'], constants.CONTENT_TYPE_JSON)


    @patch('ztpserver.controller.create_repository')
    def test_empty_xmpp(self, m_repository):
        cfg = {'return_value.read.return_value': {'logging': [],
                                                  'xmpp': None}}
        m_repository.return_value.get_file.configure_mock(**cfg)

        controller = ztpserver.controller.BootstrapController()

        request = Request.blank('')
        request.remote_addr = ''
        resp = controller.config(request)

        self.assertEqual(resp['body'], controller.DEFAULT_CONFIG)
        self.assertEqual(resp['content_type'], constants.CONTENT_TYPE_JSON)

    @patch('ztpserver.controller.create_repository')
    def test_empty_logging(self, m_repository):
        cfg = {'return_value.read.return_value': {'logging': None,
                                                  'xmpp': {}}}
        m_repository.return_value.get_file.configure_mock(**cfg)

        controller = ztpserver.controller.BootstrapController()

        request = Request.blank('')
        request.remote_addr = ''
        resp = controller.config(request)

        self.assertEqual(resp['body'], controller.DEFAULT_CONFIG)
        self.assertEqual(resp['content_type'], constants.CONTENT_TYPE_JSON)

    @patch('ztpserver.controller.create_repository')
    def test_empty_xmpp_logging(self, m_repository):
        cfg = {'return_value.read.return_value': {'logging': None,
                                                  'xmpp': None}}
        m_repository.return_value.get_file.configure_mock(**cfg)

        controller = ztpserver.controller.BootstrapController()

        request = Request.blank('')
        request.remote_addr = ''
        resp = controller.config(request)

        self.assertEqual(resp['body'], controller.DEFAULT_CONFIG)
        self.assertEqual(resp['content_type'], constants.CONTENT_TYPE_JSON)


class BootstrapUnitTests(unittest.TestCase):

    def setUp(self):
        self.m_repository = Mock()
        ztpserver.controller.create_repository = self.m_repository

    @patch('string.Template.safe_substitute')
    def test_get_bootstrap_success(self, m_substitute):
        contents = random_string()
        m_substitute.return_value = contents

        request = Request.blank('/bootstrap', headers=ztp_headers())
        resp = request.get_response(ztpserver.controller.Router())

        self.assertEqual(resp.status_code, constants.HTTP_STATUS_OK)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_PYTHON)
        self.assertEqual(resp.body, contents)

    def test_get_bootstrap_missing(self):
        cfg = {'return_value.get_file.side_effect': FileObjectNotFound}
        self.m_repository.configure_mock(**cfg)

        request = Request.blank('/bootstrap', headers=ztp_headers())
        resp = request.get_response(ztpserver.controller.Router())

        self.assertEqual(resp.status_code, constants.HTTP_STATUS_BAD_REQUEST)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_HTML)

    @patch('string.Template.safe_substitute')
    def test_get_bootstrap_misconfigured(self, m_substitute):
        m_substitute.side_effect = KeyError

        request = Request.blank('/bootstrap', headers=ztp_headers())
        resp = request.get_response(ztpserver.controller.Router())

        self.assertEqual(resp.status_code, constants.HTTP_STATUS_BAD_REQUEST)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_HTML)

    def test_get_bootstrap_inaccessible(self):
        cfg = {'return_value.get_file.side_effect': FileObjectError}
        self.m_repository.configure_mock(**cfg)

        request = Request.blank('/bootstrap', headers=ztp_headers())
        resp = request.get_response(ztpserver.controller.Router())

        self.assertEqual(resp.status_code, constants.HTTP_STATUS_BAD_REQUEST)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_HTML)

    def test_get_bootstrap_config_success(self):
        config = create_bootstrap_conf()
        config.add_logging(dict(destination=random_string(),
                                level=random_string()))

        cfg = {'return_value.read.return_value': config.as_dict()}
        self.m_repository.return_value.get_file.configure_mock(**cfg)

        request = Request.blank('/bootstrap/config')
        resp = request.get_response(ztpserver.controller.Router())

        self.assertEqual(resp.status_code, constants.HTTP_STATUS_OK)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_JSON)
        self.assertEqual(json.loads(resp.body), config.as_dict())

    def test_get_bootstrap_config_defaults(self):
        cfg = {'return_value.get_file.side_effect': FileObjectNotFound}
        self.m_repository.configure_mock(**cfg)

        request = Request.blank('/bootstrap/config')
        resp = request.get_response(ztpserver.controller.Router())

        defaultconfig = {'logging': list(), 'xmpp': dict()}

        self.assertEqual(resp.status_code, constants.HTTP_STATUS_OK)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_JSON)
        self.assertEqual(json.loads(resp.body), defaultconfig)


class FilesControllerIntegrationTests(unittest.TestCase):

    def tearDown(self):
        remove_all()

    @patch('ztpserver.controller.create_repository')
    def test_get_file_success(self, m_repository):
        contents = random_string()
        filepath = write_file(contents)

        m_repository.return_value.get_file.return_value.name = filepath

        url = '/files/%s' % filepath
        request = Request.blank(url)
        resp = request.get_response(ztpserver.controller.Router())

        m_repository.return_value.get_file.assert_called_with(filepath)
        self.assertEqual(resp.status_code, constants.HTTP_STATUS_OK)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_OTHER)
        self.assertEqual(resp.body, contents)


    @patch('ztpserver.controller.create_repository')
    def test_get_missing_file(self, m_repository):
        cfg = {'return_value.get_file.side_effect':
                    ztpserver.repository.FileObjectNotFound}
        m_repository.configure_mock(**cfg)

        filename = random_string()
        url = '/files/%s' % filename

        request = Request.blank(url)
        resp = request.get_response(ztpserver.controller.Router())

        filepath = 'files/%s' % filename
        m_repository.return_value.get_file.assert_called_with(filepath)
        self.assertEqual(resp.status_code, constants.HTTP_STATUS_NOT_FOUND)


class ActionsControllerIntegrationTests(unittest.TestCase):

    @patch('ztpserver.controller.create_repository')
    def test_get_action_success(self, m_repository):
        contents = random_string()
        cfg = {'return_value.read.return_value': contents}
        m_repository.return_value.get_file.configure_mock(**cfg)

        filename = random_string()
        url = '/actions/%s' % filename

        request = Request.blank(url)
        resp = request.get_response(ztpserver.controller.Router())

        m_repository.return_value.get_file.assert_called_with(url[1:])
        self.assertEqual(resp.status_code, constants.HTTP_STATUS_OK)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_PYTHON)
        self.assertEqual(resp.body, contents)

    @patch('ztpserver.controller.create_repository')
    def test_get_action_missing(self, m_repository):
        cfg = {'return_value.get_file.side_effect': FileObjectNotFound}
        m_repository.configure_mock(**cfg)

        filename = random_string()
        url = '/actions/%s' % filename

        request = Request.blank(url)
        resp = request.get_response(ztpserver.controller.Router())

        self.assertEqual(resp.status_code, constants.HTTP_STATUS_NOT_FOUND)


class NodesControllerUnitTests(unittest.TestCase):


    def tearDown(self):
        ztpserver.config.runtime.set_value(\
            'disable_topology_validation', False, 'default')
        ztpserver.config.runtime.set_value(\
            'identifier', 'serialnumber', 'default')
        
    @classmethod
    def identifier(cls, node):
        identifier = ztpserver.config.runtime.default.identifier
        if identifier == 'systemmac':
            return node.systemmac
        else:
            return node.serialnumber

    @patch('ztpserver.controller.create_repository')
    def test_create(self, m_repository):
        node = Mock(systemmac=random_string(), serialnumber=random_string())
        body = dict(systemmac=node.systemmac, serialnumber=node.serialnumber)

        request = Request.blank('/nodes')
        request.body = json.dumps(body)

        controller = ztpserver.controller.NodesController()
        with patch.object(controller, 'fsm') as m_fsm:
            controller.create(request)
            self.assertEqual(self.identifier(node),
                             m_fsm.call_args[1]['node_id'])

    @patch('ztpserver.controller.create_repository')
    def test_create_systemmac(self, m_repository):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_create()

    @patch('ztpserver.controller.create_repository')
    def test_create_missing_identifier(self, m_repository):
        node = Mock(systemmac=None, serialnumber=None)
        body = dict(systemmac=node.systemmac, serialnumber=node.serialnumber)

        request = Request.blank('/nodes')
        request.body = json.dumps(body)

        controller = ztpserver.controller.NodesController()
        resp = controller.create(request)
        self.assertEqual(resp.status_code, constants.HTTP_STATUS_BAD_REQUEST)

    @patch('ztpserver.controller.create_repository')
    def test_create_missing_identifier_systemmac(self, m_repository):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_create_missing_identifier()

    @patch('ztpserver.controller.create_repository')
    def test_node_exists(self, m_repository):
        m_repository.return_value.exists.return_value = True

        node = Mock(serialnumber=random_string(),
                    systemmac=random_string())

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.node_exists(dict(), node=node,
                                               node_id=self.identifier(node))

        self.assertEqual(state, 'dump_node')
        self.assertEqual(resp['status'], constants.HTTP_STATUS_CONFLICT)

    @patch('ztpserver.controller.create_repository')
    def test_node_exists_systemmac(self, m_repository):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_node_exists()


    @patch('ztpserver.controller.create_repository')
    def test_node_exists_definition_exists(self, m_repository):
        node = create_node()
        cfg = dict()

        def m_exists(arg):
            if arg.endswith(DEFINITION_FN):
                return True
            return False
        cfg['return_value.exists.side_effect'] = m_exists

        m_repository.configure_mock(**cfg)

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.node_exists(dict(), node=node,
                                               node_id=self.identifier(node))

        self.assertEqual(state, 'dump_node')
        self.assertEqual(resp['status'], constants.HTTP_STATUS_CONFLICT)

    @patch('ztpserver.controller.create_repository')
    def test_node_exists_definition_exists_systemmac(self, m_repository):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_node_exists_definition_exists()

    @patch('ztpserver.controller.create_repository')
    def test_node_exists_startup_config_exists(self, m_repository):
        node = create_node()
        cfg = dict()

        def m_exists(arg):
            if arg.endswith(ztpserver.controller.STARTUP_CONFIG_FN):
                return True
            return False
        cfg['return_value.exists.side_effect'] = m_exists

        m_repository.configure_mock(**cfg)

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.node_exists(dict(), node=node,
                                               node_id=self.identifier(node))

        self.assertEqual(state, 'dump_node')
        self.assertEqual(resp['status'], constants.HTTP_STATUS_CONFLICT)

    @patch('ztpserver.controller.create_repository')
    def test_node_exists_startup_config_exists_systemmac(self, m_repository):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_node_exists_startup_config_exists()

    @patch('ztpserver.controller.create_repository')
    def test_node_exists_sysmac_folder_exists(self, m_repository):
        node = create_node()
        cfg = dict()

        def m_exists(arg):
            if arg.endswith(self.identifier(node)):
                return True
            return False
        cfg['return_value.exists.side_effect'] = m_exists

        m_repository.configure_mock(**cfg)

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.node_exists(dict(), node=node,
                                               node_id=self.identifier(node))

        self.assertEqual(state, None)
        self.assertEqual(resp['status'],
                         constants.HTTP_STATUS_BAD_REQUEST)

    @patch('ztpserver.controller.create_repository')
    def test_node_exists_sysmac_folder_exists_systemmac(self, m_repository):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_node_exists_sysmac_folder_exists()

    @patch('ztpserver.controller.create_repository')
    def test_node_exists_failure(self, m_repository):
        m_repository.return_value.exists.return_value = False

        node = Mock(serialnumber=random_string(),
                    systemmac=random_string())
        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.node_exists(dict(), node=node,
                                               node_id=self.identifier(node))

        self.assertEqual(state, 'post_config')
        self.assertIsInstance(resp, dict)
        self.assertEqual(resp, dict())

    @patch('ztpserver.controller.create_repository')
    def test_node_exists_failure_systemmac(self, m_repository):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_node_exists_failure()

    @patch('ztpserver.controller.create_repository')
    def test_dump_node_success(self, m_repository):
        node = Mock(serialnumber=random_string(),
                    systemmac=random_string())

        cfg = dict()
        cfg['return_value.get_file'] = Mock()
        m_repository.configure_mock(**cfg)

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.dump_node(dict(), node=node,
                                             node_id=self.identifier(node))

        self.assertEqual(state, 'set_location')
        self.assertIsInstance(resp, dict)
        self.assertEqual(resp, dict())

    @patch('ztpserver.controller.create_repository')
    def test_dump_node_success_systemmac(self, m_repository):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_dump_node_success()

    @patch('ztpserver.controller.create_repository')
    def test_dump_node_write_file_failure(self, m_repository):

        cfg = {'return_value.get_file.return_value.write.side_effect': \
               ztpserver.repository.FileObjectError}
        m_repository.configure_mock(**cfg)

        node = Mock(serialnumber=random_string(),
                    systemmac=random_string())

        controller = ztpserver.controller.NodesController()
        self.assertRaises(ztpserver.repository.FileObjectError,
                          controller.dump_node,
                          dict(),
                          node=node,
                          node_id=self.identifier(node))

    @patch('ztpserver.controller.create_repository')
    def test_dump_node_write_file_failure_systemmac(self, m_repository):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_dump_node_write_file_failure()

    def test_set_location_success(self):
        node = Mock(serialnumber=random_string(),
                    systemmac=random_string())
        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.set_location(dict(), node=node,
                                                node_id=self.identifier(node))

        location = 'nodes/%s' % self.identifier(node)
        self.assertIsNone(state)
        self.assertIsInstance(resp, dict)
        self.assertEqual(resp['location'], location)

    def test_set_location_failure(self):
        controller = ztpserver.controller.NodesController()
        self.assertRaises(AttributeError, controller.set_location,
                          dict(), node=None)

    def test_post_config_success(self):
        request = Mock(json=dict(config=random_string()))
        node = Mock(serialnumber=random_string(),
                    systemmac=random_string())

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.post_config(dict(), request=request,
                                               node=node,
                                               node_id=self.identifier(node))

        self.assertEqual(state, 'set_location')
        self.assertIsInstance(resp, dict)
        self.assertEqual(resp['status'], constants.HTTP_STATUS_CREATED)

    def test_post_config_success_systemmac(self):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_post_config_success()

    def test_post_config_key_error_failure(self):
        request = Mock(json=dict())
        node = Mock(serialnumber=random_string(),
                    systemmac=random_string())

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.post_config(dict(), request=request,
                                               node=node,
                                               node_id=self.identifier(node))

        self.assertEqual(state, 'post_node')
        self.assertIsInstance(resp, dict)
        self.assertEqual(resp, dict())

    def test_post_config_key_error_failure_systemmac(self):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_post_config_key_error_failure()

    def test_post_config_failure(self):
        controller = ztpserver.controller.NodesController()
        self.assertRaises(Exception, controller.post_config, dict())

    @patch('ztpserver.controller.load_neighbordb')
    def test_post_node_success_single_match(self, m_load_neighbordb):
        request = Mock(json=dict(neighbors=dict()))
        node = Mock(serialnumber=random_string(),
                    systemmac=random_string())

        m_load_neighbordb.return_value.match_node.return_value = [mock_match()]
        controller = ztpserver.controller.NodesController()

        (resp, state) = controller.post_node(dict(), request=request, node=node,
                                             node_id=self.identifier(node))

        self.assertEqual(state, 'dump_node')
        self.assertIsInstance(resp, dict)
        self.assertEqual(resp['status'], constants.HTTP_STATUS_CREATED)

    @patch('ztpserver.controller.load_neighbordb')
    def test_post_node_success_single_match_systemmac(self, m_load_neighbordb):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_post_node_success_single_match()

    @patch('ztpserver.controller.load_neighbordb')
    def test_post_node_success_multiple_matches(self, m_load_neighbordb):
        request = Mock(json=dict(neighbors=dict()))
        node = Mock(serialnumber=random_string(),
                    systemmac=random_string())

        m_load_neighbordb.return_value.match_node.return_value = [mock_match(),
                                                                  mock_match(),
                                                                  mock_match()]

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.post_node(dict(), request=request, node=node,
                                             node_id=self.identifier(node))

        self.assertEqual(state, 'dump_node')
        self.assertIsInstance(resp, dict)
        self.assertEqual(resp['status'], constants.HTTP_STATUS_CREATED)

    @patch('ztpserver.controller.load_neighbordb')
    def test_post_node_success_multiple_matches_systemmac(self,
                                                          m_load_neighbordb):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_post_node_success_multiple_matches()

    @patch('ztpserver.controller.load_neighbordb')
    def test_post_node_failure_no_matches(self, m_load_neighbordb):
        request = Mock(json=dict(neighbors=dict()))
        node = Mock(serialnumber=random_string(),
                    systemmac=random_string())

        m_load_neighbordb.return_value.match_node.return_value = list()

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.post_node(dict(), request=request,
                                             node=node,
                                             node_id=self.identifier(node))
        self.assertEqual(state, None)
        self.assertIsInstance(resp, dict)
        self.assertEqual(resp['status'], constants.HTTP_STATUS_BAD_REQUEST)

    @patch('ztpserver.controller.load_neighbordb')
    def test_post_node_failure_no_matches_systemac(self, m_load_neighbordb):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_post_node_failure_no_matches()

    @patch('ztpserver.controller.load_neighbordb')
    def test_post_node_no_definition_in_pattern(self, m_load_neighbordb):
        request = Mock(json=dict(neighbors=dict()))
        node = Mock(serialnumber=random_string(),
                    systemmac=random_string())

        pattern = Mock()
        del pattern.definition

        m_load_neighbordb.return_value.match_node.return_value = [pattern]

        controller = ztpserver.controller.NodesController()
        self.assertRaises(AttributeError, controller.post_node, dict(),
                          request=request, node=node,
                          node_id=self.identifier(node))

    @patch('ztpserver.controller.load_neighbordb')
    def test_post_node_no_definition_in_pattern_systemmac(self,
                                                          m_load_neighbordb):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_post_node_no_definition_in_pattern()

    @patch('ztpserver.controller.create_repository')
    def test_get_definition_success(self, m_repository):
        cfg = {'return_value.read.return_value': MagicMock()}
        m_repository.return_value.get_file.configure_mock(**cfg)

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.get_definition(dict(),
                                                  resource=random_string())

        self.assertEqual(state, 'get_attributes')
        self.assertIsInstance(resp, dict)

    @patch('ztpserver.controller.replace_config_action')
    def test_get_startup_config_success(self, m_replace_config_action):
        m_replace_config_action.return_value = dict()

        ztpserver.topology.replace_config_action = Mock(return_value=dict())

        response = dict(definition={'actions': list()})

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.get_startup_config(response,
                                                      resource=random_string())

        self.assertEqual(state, 'get_definition')
        self.assertIsInstance(resp, dict)


    @patch('ztpserver.controller.replace_config_action')
    def test_get_startup_config_success_no_definition(self,
                                                      m_replace_config_action):
        resource = random_string()

        action_name = random_string()
        action = {'name': action_name, 'action': 'replace_config'}
        m_replace_config_action.return_value = action
        ztpserver.topology.replace_config_action = Mock(return_value=action)

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.get_startup_config(dict(), resource=resource)

        self.assertEqual(state, 'get_definition')
        self.assertIsInstance(resp, dict)
        self.assertTrue(resp['get_startup_config'])
        self.assertTrue(resp['definition'], 'Autogenerated definition')
        self.assertEqual(resp['definition']['actions'][0]['name'], action_name)

    @patch('ztpserver.controller.load_pattern')
    def test_do_validation_success(self, m_load_pattern):

        cfg = {'return_value.match_node.return_value': True}
        m_load_pattern.configure_mock(**cfg)

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.do_validation(dict(),
                                                 resource=random_string(),
                                                 node=Mock())

        self.assertEqual(state, 'get_startup_config')
        self.assertIsInstance(resp, dict)

    def test_do_validation_disabled_success(self):
        ztpserver.config.runtime.set_value(\
            'disable_topology_validation', True, 'default')

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.do_validation(dict(),
                                                 resource=random_string())

        self.assertEqual(state, 'get_startup_config')
        self.assertIsInstance(resp, dict)

    @patch('ztpserver.controller.create_repository')
    def test_get_attributes_success(self, m_repository):
        cfg = {'return_value.read.return_value': random_string()}
        m_repository.return_value.get_file.configure_mock(**cfg)

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.get_attributes(dict(),
                                                  resource=random_string())

        self.assertEqual(state, 'do_substitution')
        self.assertIsInstance(resp, dict)

    @patch('ztpserver.controller.create_repository')
    def test_get_attributes_missing(self, m_repository):
        node = create_node()

        m_repository.return_value.exists.return_value = False
        m_repository.return_value.get_file = \
            Mock(side_effect=ztpserver.repository.FileObjectNotFound)


        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.get_attributes(
            dict(),
            resource=self.identifier(node))

        self.assertEqual(state, 'do_substitution')
        self.assertIsInstance(resp, dict)
        self.assertEqual(resp['attributes'], dict())

    @patch('ztpserver.controller.create_repository')
    def test_get_attributes_missing_systemmac(self, m_repository):
        ztpserver.config.runtime.set_value(\
            'identifier', 'systemmac', 'default')
        self.test_get_attributes_missing()

    def test_do_substitution_success(self):
        defattrs = dict(foo='$foo')

        definition = create_definition()
        definition.add_attribute('foo', 'bar')
        definition.add_action(name='dummy action', attributes=defattrs)

        response = dict(definition=definition.as_dict())

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.do_substitution(response,
                                                   resource=random_string())

        self.assertEqual(state, 'do_resources')
        self.assertIsInstance(resp, dict)

        foo = resp['definition']['actions'][0]['attributes']['foo']
        self.assertEqual(foo, 'bar')

    def test_do_substitution_with_attributes(self):
        defattrs = dict(foo='$foo')

        definition = create_definition()
        definition.add_attribute('foo', 'bar')
        definition.add_action(name='dummy action', attributes=defattrs)

        g_attr_foo = random_string()
        attributes = create_attributes()
        attributes.add_attribute('foo', g_attr_foo)

        response = dict(definition=definition.as_dict(),
                        attributes=attributes.as_dict())

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.do_substitution(response,
                                                   resource=random_string())

        self.assertEqual(state, 'do_resources')
        self.assertIsInstance(resp, dict)

        foo = resp['definition']['actions'][0]['attributes']['foo']
        self.assertEqual(foo, g_attr_foo)

    def test_do_substitution_without_actions(self):
        # github issue #129
        definition = create_definition()
        definition.actions.append({'name': 'foo'})
        response = dict(definition=definition.as_dict())

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.do_substitution(response,
                                                   resource=random_string())

        self.assertEqual(state, 'do_resources')
        self.assertIsInstance(resp, dict)

    def test_do_resources_success(self):
        var_foo = random_string()
        ztpserver.controller.load_resources = \
            Mock(return_value=dict(foo=var_foo))

        definition = create_definition()
        definition.add_action(name='dummy action',
                              attributes=dict(foo=random_string()))

        response = dict(definition=definition.as_dict())

        controller = ztpserver.controller.NodesController()
        (resp, state) = controller.do_resources(response, node=Mock(),
                                                resource=random_string())

        self.assertEqual(state, 'finalize_response')
        self.assertIsInstance(resp, dict)
        foo = resp['definition']['actions'][0]['attributes']['foo']
        self.assertEqual(foo, var_foo)

    @patch('os.path.isfile')
    def test_put_config_success(self, m_is_file):
        m_is_file.return_value = False

        resource = random_string()
        body = random_string()
        request = Mock(content_type=constants.CONTENT_TYPE_OTHER, body=body)

        controller = ztpserver.controller.NodesController()
        resp = controller.put_config(request,
                                     resource=resource)

        self.assertEqual(resp, dict())


class NodesControllerPostFsmIntegrationTests(unittest.TestCase):

    def setUp(self):
        ztpserver.config.runtime.set_value(\
            'disable_topology_validation', False, 'default')

    @patch('ztpserver.controller.create_repository')
    def test_missing_required_attributes(self, m_repository):
        url = '/nodes'
        body = json.dumps(dict())

        request = Request.blank(url, body=body, method='POST',
                                headers=ztp_headers())

        resp = request.get_response(ztpserver.controller.Router())
        self.assertEqual(resp.status_code, constants.HTTP_STATUS_BAD_REQUEST)

    @patch('ztpserver.controller.create_repository')
    def test_node_exists(self, m_repository):
        url = '/nodes'
        serialnumber = random_string()
        body = json.dumps(dict(serialnumber=serialnumber))

        m_repository.return_value.exists.return_value = True

        request = Request.blank(url, body=body, method='POST',
                                headers=ztp_headers())

        resp = request.get_response(ztpserver.controller.Router())

        location = 'http://localhost/nodes/%s' % serialnumber
        self.assertEqual(resp.status_code, constants.HTTP_STATUS_CONFLICT)
        self.assertEqual(resp.location, location)

    @patch('ztpserver.controller.create_repository')
    def test_post_config(self, m_repository):
        url = '/nodes'
        serialnumber = random_string()
        config = random_string()
        body = json.dumps(dict(serialnumber=serialnumber, config=config))

        m_repository.return_value.exists.return_value = False

        request = Request.blank(url, body=body, method='POST',
                                headers=ztp_headers())
        resp = request.get_response(ztpserver.controller.Router())

        location = 'http://localhost/nodes/%s' % serialnumber
        self.assertEqual(resp.status_code, constants.HTTP_STATUS_CREATED)
        self.assertEqual(resp.location, location)

    @patch('ztpserver.controller.create_repository')
    @patch('ztpserver.controller.load_neighbordb')
    def test_post_node_success(self, m_load_neighbordb, m_repository):
        node = create_node()

        definition = create_definition()
        definition.add_action()

        cfg = {'return_value.exists.return_value': False}
        m_repository.configure_mock(**cfg)

        pattern_name = random_string()
        cfg = {'return_value.match_node.return_value':
               [mock_match(name=pattern_name)]}
        m_load_neighbordb.configure_mock(**cfg)

        request = Request.blank('/nodes', body=node.as_json(), method='POST',
                                headers=ztp_headers())
        resp = request.get_response(ztpserver.controller.Router())

        args_list = list()
        args_list.append('nodes/%s/%s' % (node.serialnumber, DEFINITION_FN))
        args_list.append('nodes/%s/%s' % (node.serialnumber, PATTERN_FN))

        for arg in args_list:
            m_repository.return_value.add_file.assert_any_call(arg)

        write_mock = m_repository.return_value.add_file.return_value.write
        # 'definition' is not written to the pattern file
        # Empty 'variables', 'node' are not written to the
        # pattern file either
        self.assertEqual(sorted(write_mock.call_args_list[1][0][0].keys()),
                         ['interfaces', 'name'])

        location = 'http://localhost/nodes/%s' % node.serialnumber
        self.assertEqual(resp.status_code, constants.HTTP_STATUS_CREATED)
        self.assertEqual(resp.location, location)


class NodesControllerGetFsmIntegrationTests(unittest.TestCase):

    def setUp(self):
        ztpserver.config.runtime.set_value(\
            'disable_topology_validation', False, 'default')

    @patch('ztpserver.controller.create_repository')
    def test_get_fsm_missing_node(self, m_repository):
        cfg = {'return_value.get_file.side_effect': FileObjectNotFound}
        m_repository.configure_mock(**cfg)

        url = '/nodes/%s' % random_string()
        request = Request.blank(url, method='GET')
        resp = request.get_response(ztpserver.controller.Router())

        self.assertEqual(resp.status_code, constants.HTTP_STATUS_BAD_REQUEST)


    @patch('ztpserver.controller.create_node')
    @patch('ztpserver.controller.create_repository')
    @patch('ztpserver.controller.load_pattern')
    def test_get_startup_config_wo_validation(self, m_load_pattern,
                                              m_repository, m_create_node):

        ztpserver.config.runtime.set_value(\
            'disable_topology_validation', True, 'default')

        definition = create_definition()
        definition.add_action()

        node = create_node()
        cfg = dict()

        def m_get_file(arg):
            fileobj = Mock()
            if arg.endswith('.node'):
                fileobj.return_value.read.return_value = node.as_dict()
            elif arg.endswith('startup-config'):
                fileobj.return_value.read.return_value = random_string()
            else:
                raise ztpserver.repository.FileObjectNotFound
            return fileobj
        cfg['return_value.get_file'] = Mock(side_effect=m_get_file)

        m_repository.configure_mock(**cfg)

        url = '/nodes/%s' % node.serialnumber
        request = Request.blank(url, method='GET')
        resp = request.get_response(ztpserver.controller.Router())

        self.assertEqual(resp.status_code, constants.HTTP_STATUS_OK)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_JSON)
        self.assertIsInstance(json.loads(resp.body), dict)

    @patch('ztpserver.controller.load_pattern')
    @patch('ztpserver.controller.create_repository')
    @patch('ztpserver.controller.create_node')
    def test_get_startup_config_w_validation_success(self,
                                                     m_load_pattern,
                                                     m_repository,
                                                     m_create_node):

        definition = create_definition()
        definition.add_action()

        node = create_node()
        cfg = dict()

        def m_get_file(arg):
            fileobj = Mock()
            if arg.endswith('.node'):
                fileobj.read.return_value = node.as_dict()
            elif arg.endswith('startup-config'):
                fileobj.read.return_value = random_string()
            elif arg.endswith('pattern'):
                fileobj.read.return_value = random_string()
            else:
                raise ztpserver.repository.FileObjectNotFound
            return fileobj

        cfg['return_value.get_file'] = Mock(side_effect=m_get_file)

        m_repository.configure_mock(**cfg)

        cfg = {'return_value.match_node.return_value': True}
        m_load_pattern.configure_mock(**cfg)

        url = '/nodes/%s' % node.serialnumber
        request = Request.blank(url, method='GET')
        resp = request.get_response(ztpserver.controller.Router())

        self.assertEqual(resp.status_code, constants.HTTP_STATUS_OK)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_JSON)
        self.assertIsInstance(json.loads(resp.body), dict)

    @patch('ztpserver.controller.create_repository')
    @patch('ztpserver.controller.load_pattern')
    def test_get_startup_config_w_validation_failure(self, m_load_pattern,
                                                     m_repository):

        m_load_pattern.return_value.match_node.return_value = list()

        node = create_node()
        cfg = dict()

        def exists(filepath):
            if filepath.endswith('startup-config'):
                return True
            return False
        cfg['return_value.exists'] = Mock(side_effect=exists)

        def get_file(filepath):
            fileobj = Mock()
            if filepath.endswith('node'):
                fileobj.contents = node.as_json()
                return fileobj
            elif filepath.endswith('pattern'):
                return fileobj
            raise ztpserver.repository.FileObjectNotFound
        cfg['return_value.get_file'] = Mock(side_effect=get_file)

        m_repository.configure_mock(**cfg)

        url = '/nodes/%s' % node.serialnumber
        request = Request.blank(url, method='GET')
        resp = request.get_response(ztpserver.controller.Router())

        self.assertEqual(resp.status_code, constants.HTTP_STATUS_BAD_REQUEST)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_HTML)
        self.assertEqual(resp.body, str())

    @patch('ztpserver.controller.create_repository')
    def test_get_definition_wo_validation(self, m_repository):
        ztpserver.config.runtime.set_value(\
            'disable_topology_validation', True, 'default')

        definitions_file = create_definition()
        definitions_file.add_action()

        node = create_node()
        cfg = dict()

        def m_get_file(arg):
            m_file_object = Mock()
            if arg.endswith('.node'):
                m_file_object.read.return_value = node.as_dict()
            elif arg.endswith('definition'):
                m_file_object.read.return_value = definitions_file.as_dict()
            else:
                raise ztpserver.repository.FileObjectNotFound
            return m_file_object

        cfg['return_value.get_file.side_effect'] = m_get_file
        m_repository.configure_mock(**cfg)

        url = '/nodes/%s' % node.serialnumber
        request = Request.blank(url, method='GET')
        resp = request.get_response(ztpserver.controller.Router())

        self.assertEqual(resp.status_code, constants.HTTP_STATUS_OK)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_JSON)
        self.assertIsInstance(json.loads(resp.body), dict)

    @patch('ztpserver.controller.create_repository')
    @patch('ztpserver.controller.load_pattern')
    def test_get_definition_w_validation_success(self, m_load_pattern,
                                                 m_repository):
        node = create_node()

        definitions_file = create_definition()
        definitions_file.add_action()

        cfg = dict()

        def m_get_file(arg):
            m_file_object = Mock()
            if arg.endswith('.node'):
                m_file_object.read.return_value = node.as_dict()
            elif arg.endswith('definition'):
                m_file_object.read.return_value = definitions_file.as_dict()
            elif arg.endswith('attributes'):
                raise ztpserver.repository.FileObjectNotFound
            return m_file_object

        cfg['return_value.get_file.side_effect'] = m_get_file

        m_repository.configure_mock(**cfg)

        m_load_pattern.return_value.match_node.return_value = Mock()

        url = '/nodes/%s' % node.serialnumber
        request = Request.blank(url, method='GET')
        resp = request.get_response(ztpserver.controller.Router())

        self.assertEqual(resp.status_code, constants.HTTP_STATUS_OK)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_JSON)
        self.assertIsInstance(json.loads(resp.body), dict)


    @patch('ztpserver.controller.create_repository')
    @patch('ztpserver.controller.load_pattern')
    def test_get_definition_w_validation_failure(self, m_load_pattern,
                                                 m_repository):
        definitions_file = create_definition()
        definitions_file.add_action()

        serialnumber = random_string()
        node = Mock(serialnumber=serialnumber)
        node.identifier.return_value = serialnumber

        cfg = dict()

        def m_get_file(arg):
            m_file_object = Mock()
            if arg.endswith('.node'):
                m_file_object.read.return_value = node
            elif arg.endswith('definition'):
                m_file_object.read.return_value = definitions_file.as_dict()
            return m_file_object

        cfg['return_value.get_file.side_effect'] = m_get_file
        m_repository.configure_mock(**cfg)

        m_load_pattern.return_value.match_node.return_value = False

        url = '/nodes/%s' % node.serialnumber
        request = Request.blank(url, method='GET')
        resp = request.get_response(ztpserver.controller.Router())

        self.assertEqual(resp.status_code, constants.HTTP_STATUS_BAD_REQUEST)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_HTML)
        self.assertEqual(resp.body, str())

    @patch('ztpserver.controller.create_repository')
    @patch('ztpserver.controller.load_pattern')
    def test_get_definition_w_attributes_no_substitution(self, m_load_pattern,
                                                         m_repository):

        node = create_node()

        g_attr_foo = random_string()
        attributes_file = create_attributes()
        attributes_file.add_attribute('variables', {'foo': g_attr_foo})

        l_attr_url = random_string()
        definitions_file = create_definition()
        definitions_file.add_attribute('foo', random_string())
        definitions_file.add_action(name='dummy action',
                                    attributes=dict(url=l_attr_url))

        cfg = dict()
        def m_get_file(arg):
            m_file_object = Mock()
            if arg.endswith('.node'):
                m_file_object.read.return_value = node.as_dict()
            elif arg.endswith('definition'):
                m_file_object.read.return_value = definitions_file.as_dict()
            elif arg.endswith('attributes'):
                m_file_object.read.return_value = attributes_file.as_dict()
            elif arg.endswith('startup-config'):
                raise ztpserver.repository.FileObjectNotFound
            return m_file_object
        cfg['return_value.get_file'] = Mock(side_effect=m_get_file)

        cfg['return_value.match_node.return_value'] = True
        m_repository.configure_mock(**cfg)

        url = '/nodes/%s' % node.serialnumber
        request = Request.blank(url, method='GET')

        resp = request.get_response(ztpserver.controller.Router())

        self.assertEqual(resp.status_code, constants.HTTP_STATUS_OK)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_JSON)

        attrs = resp.json['actions'][0]['attributes']
        self.assertFalse('variables' in attrs)
        self.assertFalse('foo' in attrs)
        self.assertEqual(attrs['url'], l_attr_url)

class DefinitionStartupConfigTests(unittest.TestCase):

    @patch('ztpserver.controller.create_repository')
    @patch('ztpserver.controller.load_pattern')
    def test_get_definition_w_startuo_config(self, m_load_pattern,
                                             m_repository):

        node = create_node()

        action_name_1 = random_string()
        action_name_2 = random_string()
        definitions_file = create_definition()
        definitions_file.add_action(action=action_name_1)
        definitions_file.add_action(action=action_name_2,
                                    always_execute=True)

        cfg = dict()

        def m_get_file(arg):
            m_file_object = Mock()
            if arg.endswith('.node'):
                m_file_object.read.return_value = dict(node.as_dict())
            elif arg.endswith('definition'):
                m_file_object.read.return_value = \
                    dict(definitions_file.as_dict())
            elif arg.endswith('attributes'):
                raise ztpserver.repository.FileObjectNotFound
            return m_file_object

        cfg['return_value.get_file.side_effect'] = m_get_file
        m_repository.configure_mock(**cfg)
        m_load_pattern.return_value.match_node.return_value = Mock()

        url = '/nodes/%s' % node.serialnumber
        request = Request.blank(url, method='GET')

        resp = request.get_response(ztpserver.controller.Router())
        self.assertEqual(resp.status_code, constants.HTTP_STATUS_OK)
        self.assertEqual(resp.content_type, constants.CONTENT_TYPE_JSON)
        actions = [x['action'] for x in json.loads(resp.body)['actions']]
        self.assertEqual(actions, ['replace_config',
                                   action_name_2])


if __name__ == '__main__':
    enable_logging()
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

import sqlite3 as lite
import sys

con = lite.connect('/usr/share/ztpserver/db/resources.db')

with con:

    for table in ['mgmt_subnet', 'tor_hostnames', 'ip_vlan100', 'ip_loopback']:
        print "Working on: ",table
        cur = con.cursor()
        sql = "DROP TABLE IF EXISTS `%s`" % table
        cur.execute(sql)
        sql = "CREATE TABLE `%s`(key TEXT, node_id TEXT)" % table
        cur.execute(sql)
        if table == "mgmt_subnet":
            base = "172.16.130."
            subnet = "/24"
        elif table == "tor_hostnames":
            base = "veos-dc1-pod1-tor"
            subnet = ""
        elif table == "ip_vlan100":
            base = "10.100.1."
            subnet = "/24"
        elif table == "ip_loopback":
            base = "1.1.1."
            subnet = "/32"

        for x in range(1,500):
            sql = "INSERT INTO %s VALUES('%s%s%s',NULL)" % (table, base, str(x), subnet)
            cur.execute(sql)

        sql = "SELECT * FROM `%s`" % table
        print sql
        cur.execute(sql)
        rows = cur.fetchall()
        for row in rows:
            print row



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# pylint: disable=C0103

import logging
import re
import os

from urlparse import urlsplit, urlunsplit

log = logging.getLogger(__name__)

def atoi(text):
    return int(text) if text.isdigit() else text

def natural_keys(text):
    return [atoi(c) for c in re.split(r'(\d+)', text)]

ETHERNET_RE = re.compile(r'^(e(t(h(ernet)?)?)?)(\d+)(\/(\d+)){0,2}$')
MANAGEMENT_RE = re.compile(r'^(m(a(nagement)?)?)(\d+)(\/(\d+)){0,2}$')
INTERFACE_NO_RE = re.compile(r'^(\d+)(\/(\d+)){0,2}$')
def expand_range(interfaces):
    ''' Returns a naturally sorted list of items expanded from interfaces. '''

    # pylint: disable=R0914,R0912
    
    items = set()
    prefix = None
    for group in [x.strip() for x in interfaces.split(',')]:
        ranges = [x.strip() for x in group.split('-')]
        if len(ranges) == 1:
            interface = ranges[0].lower()
            match = ETHERNET_RE.match(interface)
            if match:
                intf_no = interface[len(match.groups()[0]):]
                
                for token in intf_no.split('/'):
                    if int(token) < 1:
                        log.warning('Unable to expand '
                                    'interface range: %s '
                                    '(invalid interface number)' % 
                                    group)
                        raise TypeError('Unable to expand '
                                        'interface range: %s '
                                        '(invalid interface number)' % 
                                        group)

                prefix = 'Ethernet'
                items.add('%s%s' % (prefix, intf_no))
            else:
                match = MANAGEMENT_RE.match(interface)
                if match:
                    intf_no = interface[len(match.groups()[0]):]

                    for token in intf_no.split('/'):
                        if int(token) < 1:
                            log.warning('Unable to expand '
                                        'interface range: %s '
                                        '(invalid interface number)' % 
                                        group)
                            raise TypeError('Unable to expand '
                                            'interface range: %s '
                                            '(invalid interface number)' % 
                                            group)

                    prefix = 'Management'
                    items.add('%s%s' % (prefix, intf_no))
                else:
                    match = INTERFACE_NO_RE.match(interface)
                    if match:

                        for token in interface.split('/'):
                            if int(token) < 1:
                                log.warning('Unable to expand '
                                            'interface range: %s '
                                            '(invalid interface number)' % 
                                            group)
                                raise TypeError('Unable to expand '
                                                'interface range: %s '
                                                '(invalid interface number)' % 
                                                group)

                        items.add('%s%s' % (prefix, interface))
                    else:
                        
                        log.warning('Unable to expand interface range: %s '
                                    '(invalid interface)' % group)
                        raise TypeError('Unable to expand interface range: %s '
                                        '(invalid interface)' % group)
        elif len(ranges) == 2:
            [start, end] = [x.lower() for x in ranges]
            
            items_len = len(items)
            for regex, intf_type in [(ETHERNET_RE, 'Ethernet'),
                                     (MANAGEMENT_RE, 'Management'),
                                     (INTERFACE_NO_RE, prefix)]:
                match_start =  regex.match(start)
                if match_start:
                    if regex != INTERFACE_NO_RE:
                        start_intf_tokens = \
                            start[len(match_start.groups()[0]):].split('/')
                    else:
                        start_intf_tokens = start.split('/')

                    match_end = regex.match(end)
                    end_intf_tokens = None
                    if match_end:
                        if regex != INTERFACE_NO_RE:
                            end_intf_tokens = \
                                end[len(match_end.groups()[0]):].split('/')
                        else:
                            end_intf_tokens = end.split('/')
                    else:
                        match_end = INTERFACE_NO_RE.match(end)
                        if match_end:
                            end_intf_tokens = end.split('/')

                    if not end_intf_tokens:
                        log.warning('Unable to expand '
                                    'interface range: %s '
                                    '(invalid range end)' % group)

                        raise TypeError('Unable to expand '
                                        'interface range: %s '
                                        '(invalid range end)' % group)

                    if start_intf_tokens[:-1] != end_intf_tokens[:-1]:
                        log.warning('Unable to expand '
                                    'interface range: %s '
                                    '(invalid range)' % group)
                        raise TypeError('Unable to expand '
                                        'interface range: %s '
                                        '(invalid range)' % group)

                    start_index = int(start_intf_tokens[-1])
                    end_index = int(end_intf_tokens[-1])
                    if start_index >= end_index:
                        log.warning('Unable to expand '
                                    'interface range: %s '
                                    '(non-increasing range)' % 
                                    group)
                        raise TypeError('Unable to expand '
                                        'interface range: %s '
                                        '(non-increasing range)' % 
                                        group)

                    if start_index < 1 or end_index < 1:
                        log.warning('Unable to expand '
                                    'interface range: %s '
                                    '(invalid interface number)' % 
                                    group)
                        raise TypeError('Unable to expand '
                                        'interface range: %s '
                                        '(invalid interface number)' % 
                                        group)
                    
                    for index in range(start_index, 
                                       end_index):
                        prefix = intf_type
                        items.add('%s%s' % 
                                  (intf_type,
                                   '/'.join(start_intf_tokens[:-1] + 
                                            [str(index)])))
            if items_len == len(items):
                log.warning('Unable to expand interface range: %s ' % 
                            group)
                raise TypeError('Unable to expand interface range: %s ' % 
                                group)
        else:
            log.warning('Unable to expand interface range: %s '
                        '(invalid input)' % group)
            raise TypeError('Unable to expand interface range: %s '
                            '(invalid input)' % group)

    log.debug('%s expanded to: %s' % (interfaces, items))
    return items

def parse_interface(neighbor, node_id):
    try:
        if hasattr(neighbor, 'items'):
            remote_device = neighbor['device']
            remote_interface = neighbor.get('port', 'any')

        else:
            if neighbor == 'any':
                remote_device, remote_interface = 'any', 'any'
            elif neighbor == 'none':
                remote_device, remote_interface = 'none', 'none'
            elif ':' not in neighbor:
                remote_device = neighbor
                remote_interface = 'any'
            else:
                tokens = neighbor.split(':')
                remote_device = tokens[0]
                remote_interface = tokens[1]

        remote_device = str(remote_device).strip()
        if len(remote_device.split()) != 1:
            log.error('%s: interface parse error: invalid peer device %s' % 
                      (node_id, remote_device))
            raise Exception('%s: interface parse error: invalid peer '
                            'device %s' % (node_id, remote_device))
        
        remote_interface = str(remote_interface).strip()
        if len(remote_interface.split()) != 1:
            log.error('%s: interface parse error: invalid '
                      'peer interface %s' %
                      (node_id, remote_interface))
            raise Exception('%s: interface parse error: invalid peer '
                            'interface %s' % 
                            (node_id, remote_interface))

        return (remote_device, remote_interface)
    except KeyError as err:
        log.error('%s: interface parse error: missing attribute (%s)' % 
                  (node_id, str(err)))
        raise Exception('%s: interface parse error: missing '
                        'attribute (%s)' % (node_id, str(err)))

def url_path_join(*parts):
    """Normalize url parts and join them with a slash."""
    # pylint: disable=W0142
    schemes, netlocs, paths, queries, fragments = \
        zip(*(urlsplit(part) for part in parts))
    scheme = get_first_token(schemes)
    netloc = get_first_token(netlocs)
    path = '/'.join(x.strip('/') for x in paths if x)
    query = get_first_token(queries)
    fragment = get_first_token(fragments)
    return urlunsplit((scheme, netloc, path, query, fragment))

def get_first_token(sequence):
    return next((x for x in sequence if x), '')

def all_files(path):
    result = []
    for top, _, files in os.walk(path):
        result += [os.path.join(top, f) for f in files]
    return result



#!/usr/bin/env python
#
# Copyright (c) 2015, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#  - Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#  - Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#  - Neither the name of Arista Networks nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#pylint: disable=R0904,F0401,W0232,E1101

import os
import os.path
import unittest
import sys

sys.path.append('test/client')

from client_test_lib import Bootstrap, ActionFailureTest
from client_test_lib import eapi_log, remove_file, get_action
from client_test_lib import startup_config_action, random_string
from client_test_lib import print_action
from client_test_lib import raise_exception

class FailureTest(ActionFailureTest):

    def test_missing_url(self):
        self.basic_test('install_image',
                        'Missing attribute(\'url\')')

    def test_missing_version(self):
        self.basic_test('install_image',
                        'Missing attribute(\'version\')',
                        attributes={'url' :
                                        random_string()})

    def test_url_failure(self):
        self.basic_test('install_image',
                        'Unable to retrieve image file from URL',
                        attributes={'url' :
                                    random_string(),
                                    'version' :
                                    random_string()})


class SuccessTest(unittest.TestCase):

    def test_no_op(self):
        bootstrap = Bootstrap(ztps_default_config=True)
        version = random_string()
        bootstrap.eapi.version = version
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'test_action',
                      'attributes': {
                        'url' : random_string(),
                        'version' : version}},
                     {'action' :'startup_config_action'}])
        bootstrap.ztps.set_action_response('test_action',
                                           get_action('install_image'))
        bootstrap.ztps.set_action_response('startup_config_action',
                                           startup_config_action())
        bootstrap.start_test()

        try:
            self.failUnless(bootstrap.success())
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            bootstrap.end_test()

    def test_no_downgrade(self):
        bootstrap = Bootstrap(ztps_default_config=True)
        version1 = '4.18.1F'
        version2 = '4.17.2F'
        bootstrap.eapi.version = version1
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'test_action',
                      'attributes': {
                        'downgrade' : False,
                        'url' : random_string(),
                        'version' : version2}},
                     {'action' :'startup_config_action'}])
        bootstrap.ztps.set_action_response('test_action',
                                           get_action('install_image'))
        bootstrap.ztps.set_action_response('startup_config_action',
                                           startup_config_action())
        bootstrap.start_test()

        image_file = '%s/EOS-%s.swi' % (bootstrap.flash, version2)
        try:
            self.failUnless(bootstrap.success())
            self.failUnless('install_image: nothing to do: downgrade disabled'
                            in bootstrap.output)
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            remove_file(image_file)
            bootstrap.end_test()

    def test_success(self):
        bootstrap = Bootstrap(ztps_default_config=True)
        version = random_string()
        image = random_string()
        url = 'http://%s/%s' % (bootstrap.server, image)
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'test_action',
                      'attributes' : {
                        'url' : url,
                        'version' : version}},
                     {'action' :'startup_config_action'}])

        action = get_action('install_image')
        bootstrap.ztps.set_action_response('test_action',
                                           action)
        bootstrap.ztps.set_action_response('startup_config_action',
                                           startup_config_action())
        bootstrap.ztps.set_file_response(image, print_action())
        bootstrap.start_test()

        image_file = '%s/EOS-%s.swi' % (bootstrap.flash, version)
        try:
            self.failUnless(os.path.isfile(image_file))
            self.failUnless(bootstrap.success())
            self.failUnless(eapi_log()[-1] == 
                            'install source flash:EOS-%s.swi' % version)
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            remove_file(image_file)
            bootstrap.end_test()

if __name__ == '__main__':
    unittest.main()



# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# Copyright (c) 2015, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

CONTENT_TYPE_PYTHON = 'text/x-python'
CONTENT_TYPE_HTML = 'text/html'
CONTENT_TYPE_OTHER = 'text/plain'
CONTENT_TYPE_JSON = 'application/json'
CONTENT_TYPE_YAML = 'application/yaml'

HTTP_STATUS_OK = 200
HTTP_STATUS_CREATED = 201
HTTP_STATUS_NO_CONTENT = 204
HTTP_STATUS_BAD_REQUEST = 400
HTTP_STATUS_NOT_FOUND = 404
HTTP_STATUS_CONFLICT = 409
HTTP_STATUS_INTERNAL_SERVER_ERROR = 500



#!/usr/bin/env python
#
# Copyright (c) 2015 Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#  - Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#  - Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#  - Neither the name of Arista Networks nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#pylint: disable=R0904,F0401

import os
import os.path
import unittest

from client_test_lib import Bootstrap

from client_test_lib import fail_flash_file_action
from client_test_lib import raise_exception
from client_test_lib import random_string
from client_test_lib import startup_config_action

class BootstrapCleanupTest(unittest.TestCase):

    def test_action_failure(self):
        bootstrap = Bootstrap()
        bootstrap.ztps.set_config_response()
        bootstrap.ztps.set_node_check_response()

        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'test_action'}])

        flash_filename = random_string()
        bootstrap.ztps.set_action_response(
                'test_action',
                fail_flash_file_action(bootstrap.flash,
                                       flash_filename))

        open(bootstrap.rc_eos, 'w').write(random_string())
        open(bootstrap.startup_config, 'w').write(random_string())
        open(bootstrap.boot_extensions, 'w').write(random_string())
        os.mkdir(bootstrap.boot_extensions_folder)
        open('%s/my_extension' % 
             bootstrap.boot_extensions_folder, 'w').write(random_string())

        bootstrap.start_test()

        try:
            self.failUnless(bootstrap.eapi_node_information_collected())
            self.failUnless(bootstrap.action_failure())
            self.failIf(bootstrap.error)
            self.failIf(os.path.isfile('%s/%s' %
                                       (bootstrap.flash,
                                        flash_filename)))
            self.failIf(os.path.isfile(bootstrap.rc_eos))
            self.failIf(os.path.isfile(bootstrap.startup_config))
            self.failIf(os.path.isfile(bootstrap.boot_extensions))
            self.failIf(os.path.isdir(bootstrap.boot_extensions_folder))
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            bootstrap.end_test()

    def test_success(self):
        bootstrap = Bootstrap()

        startup_config = random_string()

        bootstrap.ztps.set_config_response()
        bootstrap.ztps.set_node_check_response()
        bootstrap.ztps.set_definition_response(
            actions=[{'action' : 'test_action'}])
        bootstrap.ztps.set_action_response(
            'test_action',
            startup_config_action([startup_config]))

        open(bootstrap.rc_eos, 'w').write(random_string())
        open(bootstrap.startup_config, 'w').write(startup_config + 
                                                  random_string())
        open(bootstrap.boot_extensions, 'w').write(random_string())
        os.mkdir(bootstrap.boot_extensions_folder)
        open('%s/my_extension' % 
             bootstrap.boot_extensions_folder, 'w').write(random_string())

        bootstrap.start_test()

        try:
            self.failUnless(bootstrap.eapi_node_information_collected())
            self.failUnless(bootstrap.success())
            self.failIf(bootstrap.error)

            self.failIf(os.path.isfile(bootstrap.rc_eos))
            self.failUnless(open(bootstrap.startup_config).read() ==
                           startup_config)
            self.failIf(os.path.isfile(bootstrap.boot_extensions))
            self.failIf(os.path.isdir(bootstrap.boot_extensions_folder))
        except AssertionError as assertion:
            print 'Output: %s' % bootstrap.output
            print 'Error: %s' % bootstrap.error
            raise_exception(assertion)
        finally:
            bootstrap.end_test()

if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# pylint: disable=W0613,C0103,R0201,W0622,W0614
#
import logging

import webob
import webob.dec
import webob.exc

from routes.middleware import RoutesMiddleware

from ztpserver.serializers import dumps
from ztpserver.constants import CONTENT_TYPE_HTML, HTTP_STATUS_OK

log = logging.getLogger(__name__)

class WSGIController(object):

    def index(self, request, **kwargs):
        return webob.exc.HTTPNoContent()

    def create(self, request, **kwargs):
        return webob.exc.HTTPNoContent()

    def new(self, request, **kwargs):
        return webob.exc.HTTPNoContent()

    def show(self, request, resource, **kwargs):
        return webob.exc.HTTPNotFound()

    def update(self, request, resource, **kwargs):
        return webob.exc.HTTPNotFound()

    def delete(self, request, resource, **kwargs):
        return webob.exc.HTTPNotFound()

    def edit(self, request, resource, **kwargs):
        return webob.exc.HTTPNotFound()

    def response(self, **kwargs):
        return webob.Response(**kwargs)

    @webob.dec.wsgify
    def __call__(self, request):
        action = request.urlvars['action']

        try:
            method = getattr(self, action)    #pylint: disable=R0921
            result = method(request, **request.urlvars)
        except Exception as exc:
            log.error('Unrecoverable error detected: %s' % exc.message)
            raise webob.exc.HTTPInternalServerError()

        if result is None:
            result = webob.exc.HTTPNoContent()

        elif isinstance(result, dict):
            # serialize body based on response content type
            if 'body' in result:
                content_type = result.get('content_type')
                result['body'] = dumps(result['body'], content_type,
                                       'general')

            result.setdefault('status', HTTP_STATUS_OK)
            result.setdefault('content_type', CONTENT_TYPE_HTML)

            result = self.response(**result)   #pylint: disable=W0142

        elif not isinstance(result, webob.Response) and \
             not isinstance(result, webob.static.FileApp):
            result = webob.exc.HTTPInternalServerError()

        return result

class WSGIRouter(object):

    def __init__(self, mapper):
        self.map = mapper
        self.router = RoutesMiddleware(self.route, self.map)

    @webob.dec.wsgify
    def __call__(self, request):
        return self.router

    @webob.dec.wsgify
    def route(self, request):
        ''' Routes the incoming request to the appropriate controller '''

        if 'controller' not in request.urlvars:
            log.debug('WSGIRouter: missing controller (request=%s)' % 
                      request)
            return webob.exc.HTTPNotFound()            

        controller = request.urlvars['controller']
        return controller()



#
# Copyright (c) 2014, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# pylint: disable=R0201
#

import collections
import logging
import json
import os
import threading
import yaml

from collections import OrderedDict

from ztpserver.constants import CONTENT_TYPE_OTHER
from ztpserver.constants import CONTENT_TYPE_JSON
from ztpserver.constants import CONTENT_TYPE_YAML

READ_WRITE_LOCK = {}
log = logging.getLogger(__name__)   #pylint: disable=C0103

class SerializerError(Exception):
    ''' base error raised by serialization functions '''
    pass


class BaseSerializer(object):
    ''' Base serializer object '''

    def __init__(self, node_id):
        self.node_id = node_id

    def serialize(self, data):
        ''' Serialize a dict to object '''
        raise NotImplementedError

    def deserialize(self, data):
        ''' Deserialize an object to dict '''
        raise NotImplementedError


class TextSerializer(BaseSerializer):

    def deserialize(self, data):
        ''' Deserialize a text object and return a dict '''
        return str(data)

    def serialize(self, data):
        ''' Serialize a dict object and return text '''
        return str(data)

#----------------------------------------------------------------------------
# Source: Michael Elsdorfer (https://gist.github.com/miracle2k))
#         https://gist.githubusercontent.com/miracle2k/
#         3184458/raw/ae89e23502f95c4555f0643dafae8a748e3fb382/
#         odict.py

def represent_odict(dump_odict, tag, mapping, flow_style=None):
    '''
    Like BaseRepresenter.represent_mapping, but does not issue the sort().
    '''
    value = []
    node = yaml.MappingNode(tag, value, flow_style=flow_style)
    if dump_odict.alias_key is not None:
        dump_odict.represented_objects[dump_odict.alias_key] = node
    best_style = True
    if hasattr(mapping, 'items'):
        mapping = mapping.items()
    for item_key, item_value in mapping:
        node_key = dump_odict.represent_data(item_key)
        node_value = dump_odict.represent_data(item_value)
        if not (isinstance(node_key, yaml.ScalarNode) and 
                not node_key.style):
            best_style = False
        if not (isinstance(node_value, yaml.ScalarNode) and 
                not node_value.style):
            best_style = False
        value.append((node_key, node_value))
    if flow_style is None:
        if dump_odict.default_flow_style is not None:
            node.flow_style = dump_odict.default_flow_style
        else:
            node.flow_style = best_style
    return node

yaml.SafeDumper.add_representer(
    OrderedDict,
    lambda dumper, 
    value: represent_odict(dumper, 
                           u'tag:yaml.org,2002:map', 
                           value))
#------------------------------------------------------------------------------

class YAMLSerializer(BaseSerializer):

    def deserialize(self, data):
        ''' Deserialize a YAML object and return a dict '''

        try:
            return yaml.safe_load(data)
        except yaml.YAMLError as err:
            msg = '''%s: unable to deserialize YAML data:
%s 

Error:
%s''' % (self.node_id, data, err)
            raise SerializerError(msg)

    def serialize(self, data):
        ''' Serialize a dict object and return YAML '''

        try:
            return yaml.safe_dump(data, default_flow_style=False)
        except yaml.YAMLError as err:
            msg = '''%s: unable to serialize YAML data:
%s 

Error:
%s''' % (self.node_id, data, err)
            raise SerializerError(msg)


class JSONSerializer(BaseSerializer):

    def deserialize(self, data):
        ''' Deserialize a JSON object and return a dict '''

        try:
            return json.loads(data)
        except Exception as err:
            msg = '''%s: unable to deserialize JSON data:
%s 

Error:
%s''' % (self.node_id, data, err)
            raise SerializerError(msg)

    def serialize(self, data):
        ''' Serialize a dict object and return JSON '''

        try:
            return json.dumps(data)
        except Exception as err:
            msg = '''%s: unable to serialize JSON data:
%s 

Error:
%s''' % (self.node_id, data, err)
            raise SerializerError(msg)


class Serializer(object):

    def __init__(self, node_id):
        self.node_id = node_id

        self._handlers = {
            CONTENT_TYPE_OTHER: TextSerializer(self.node_id),
            CONTENT_TYPE_JSON: JSONSerializer(self.node_id),
            CONTENT_TYPE_YAML: YAMLSerializer(self.node_id)
        }

    @property
    def handlers(self):
        return self._handlers

    def add_handler(self, content_type, instance):
        if content_type in self._handlers:
            log.warning('%s: overwriting previous loaded handler %s', 
                        (self.node_id, content_type))
        self._handlers[content_type] = instance

    def serialize(self, data, content_type):
        ''' Serialize the data based on the content_type '''

        handler = self.handlers.get(content_type, 
                                    TextSerializer(self.node_id))
        return handler.serialize(data)

    def deserialize(self, data, content_type=None):
        ''' Deserialize the data based on the content_type '''

        handler = self.handlers.get(content_type, 
                                    TextSerializer(self.node_id))
        data = self._convert_from_unicode(handler.deserialize(data))
        return data

    @staticmethod
    def _convert_from_unicode(data):
        if isinstance(data, basestring):
            return str(data)
        elif isinstance(data, collections.Mapping):
            return dict([Serializer._convert_from_unicode(x)
                         for x in data.items()])
        elif isinstance(data, collections.Iterable):
            return type(data)([Serializer._convert_from_unicode(x)
                               for x in data])
        else:
            return data


def loads(data, content_type, node_id):
    serializer = Serializer(node_id)
    return serializer.deserialize(data, content_type)

def load(file_path, content_type, node_id='N/A', lock=False):
    log.debug('%s: reading %s...' % (node_id, file_path))

    if lock and file_path not in READ_WRITE_LOCK:
        READ_WRITE_LOCK[file_path] = threading.Lock()

    try:
        if lock:
            with READ_WRITE_LOCK[file_path]:
                with open(file_path) as fhandler:
                    data = fhandler.read()
        else:
            with open(file_path) as fhandler:
                data = fhandler.read()            

        result = loads(data, content_type, node_id)
    except (OSError, IOError) as err:
        log.error('%s: failed to load file from %s (%s)' % 
                  (node_id, file_path, err))
        raise SerializerError('%s: failed to load file from %s (%s)' % 
                              (node_id, file_path, err))

    # Enable this log if you want to see the contents of the file (verbose)
    # log.debug('%s: loaded %s: %s' % (node_id, file_path, result))
    return result

def dumps(data, content_type, node_id):
    serializer = Serializer(node_id)
    if hasattr(data, 'serialize'):
        data = data.serialize()
    return serializer.serialize(data, content_type)


def dump(data, file_path, content_type, node_id='N/A', lock=False):
    log.debug('%s: writing %s...' % (node_id, file_path))

    if lock and file_path not in READ_WRITE_LOCK:
        READ_WRITE_LOCK[file_path] = threading.Lock()

    try:
        if lock:
            with READ_WRITE_LOCK[file_path]:
                with os.fdopen(os.open(file_path, 
                                       os.O_WRONLY | os.O_CREAT | os.O_TRUNC,
                                       0754),
                               'w') as fhandler:
                    fhandler.write(dumps(data, content_type, node_id))
        else:
            with os.fdopen(os.open(file_path, 
                                   os.O_WRONLY | os.O_CREAT | os.O_TRUNC,
                                   0754),
                           'w') as fhandler:
                fhandler.write(dumps(data, content_type, node_id))            
    except (OSError, IOError) as err:
        log.error('%s: failed to write file to %s (%s)' % 
                  (node_id, file_path, err))
        raise SerializerError('%s: failed to write file to %s (%s)' % 
                              (node_id, file_path, err))

    # Enable this log if you want to see the contents of the file (verbose)
    # log.debug('%s: wrote %s: %s' % (node_id, file_path, data))



#
# Copyright (c) 2015, Arista Networks, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
#   Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
#   Neither the name of Arista Networks nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ARISTA NETWORKS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#pylint: disable=C0103

import collections
import logging
import os
import ConfigParser

import ztpserver.types

CONF_PATH = '/etc/ztpserver'

VERSION_FILE = '.VERSION'
VERSION_FILE_PATH = '%s/%s' % (CONF_PATH, VERSION_FILE)

GLOBAL_CONF_FILE = 'ztpserver.conf'
GLOBAL_CONF_FILE_PATH = '%s/%s' % (CONF_PATH, GLOBAL_CONF_FILE)

INSTALL_PATH = '/usr/share/ztpserver'

log = logging.getLogger(__name__)

class Attr(object):
    """ Base Attribute class for deriving all attributes for a Config object

    :param name: required argument specifies attribute name
    :param type: optional keyword argument specifies attribute type.  the
                 default argument type is String
    :param group: optional keyword argument specifies attribute group.  All
                  attribute names must be unique within the group
    :param default: optional keyword argument specifies the default value for
                    the attribute.  The default value is None

    """

    def __init__(self, name, **kwargs):

        self.name = name
        self.type = kwargs.get('type') or ztpserver.types.String()
        self.group = kwargs.get('group') or 'default'
        self.default = kwargs.get('default')
        self.environ = kwargs.get('environ')

        if self.environ is not None and self.environ in os.environ:
            self.default = self.type(os.environ.get(self.environ))
        elif self.default is not None:
            self.default = self.type(self.default)

    def __repr__(self):
        return 'Attr(name=%s, group=%s, default=%s)' % \
            (self.name, self.group, self.default)


class StrAttr(Attr):
    """ String attribute class derived from Attr

    :param choices: optional keyword argument specifies valid choices
    """

    def __init__(self, name, choices=None, **kwargs):
        self.choices = choices
        attrtype = ztpserver.types.String(choices=choices)
        super(StrAttr, self).__init__(name, type=attrtype, **kwargs)


class IntAttr(Attr):
    """ Integer attribute class derived from Attr

    :param min_value: specifies the min value.  the default is None
    :param max_value: specifies the max value.  the default is None

    """

    def __init__(self, name, min_value=None, max_value=None, **kwargs):
        self.min_value = min_value
        self.max_value = max_value
        attrtype = ztpserver.types.Integer(min_value=min_value, 
                                           max_value=max_value)
        super(IntAttr, self).__init__(name, type=attrtype, **kwargs)


class BoolAttr(Attr):
    """ Boolean attribute class derived from Attr """

    def __init__(self, name, **kwargs):
        attrtype = ztpserver.types.Boolean()
        super(BoolAttr, self).__init__(name, type=attrtype, **kwargs)


class ListAttr(Attr):
    """ List attribute class derived from Attr

    :param delimiter: specifies the delimiter character to split the string on

    """

    def __init__(self, name, delimiter=',', **kwargs):
        attrtype = ztpserver.types.List(delimiter=delimiter)
        super(ListAttr, self).__init__(name, type=attrtype, **kwargs)


class Group(collections.Mapping):
    """ The Group class provides a logical grouping of attributes in a
    Config object.   Group names must be unique for each Config instance
    and cannot be assigned values.

    :param name: the name of the group
    :param config: the config object the group is associated with

    """

    def __init__(self, name, config):
        self.name = name
        self.config = config

    def __getattr__(self, name):
        # pylint: disable=W0212
        return self.config.__get_attribute__(name, self.name)

    def __getitem__(self, name):
        return self.__getattr__(name)

    def __iter__(self):
        return iter(self._keys())

    def __len__(self):
        return len(self._keys())

    def __delitem__(self):
        pass

    def __setitem__(self):
        pass

    def _keys(self):
        return [key[1] for key in self.config if key[0] == self.name]

    def add_attribute(self, item):
        self.config.add_attribute(item, self.name)


class Config(collections.Mapping):
    """ The Config class represents the configuration for collection.  """

    def __init__(self):
        self.attributes = dict()
        self.groups = list()

    def __getattr__(self, name):
        return self.__get_attribute__(name)

    def __getitem__(self, name):
        return self.__get_attribute__(name)

    def __iter__(self):
        return iter(self.attributes)

    def __len__(self):
        return len(self.attributes)

    def __repr__(self):
        return 'Config'

    def __delitem__(self):
        pass

    def __setitem__(self):
        pass

    def __get_attribute__(self, name, group=None):
        if not group and name in self.groups:
            return Group(name, self)

        key = (group, name)
        if key not in self.attributes:
            raise AttributeError('Missing attribute: %s' % str(key))

        item = self.attributes.get(key)
        return item.get('value')

    def add_attribute(self, item, group=None):

        obj = dict(_metadata=item)

        if group is None and hasattr(item, 'group'):
            group = item.group

        key = (group, item.name)

        if group not in self.groups:
            self.add_group(group)

        if key in self.attributes:
            raise AttributeError('Duplicate attribute: %s' % str(key))

        self.attributes[key] = obj
        if item.default is not None:
            obj['value'] = self._transform(obj, item.default)

    def add_group(self, group):
        if isinstance(group, Group):
            self.groups.append(group.name)
        else:
            group = str(group)
            self.groups.append(group)

    def _transform(self, item, value):
        # pylint: disable=R0201
        return item['_metadata'].type(value)

    def set_value(self, name, value, group=None):
        if not group and name in self.groups:
            raise AttributeError('Failed to set value (name=%s, group=%s): '
                                 'cannot set a value for a group' %
                                 (name, group))

        item = self.attributes.get((group, name))
        if item is None:
            raise AttributeError('Failed to set value (name=%s, group=%s): '
                                 'missing item' %
                                 (name, group))
        item['value'] = self._transform(item, value)

    def clear_value(self, name, group=None):
        """ clears the attributes value and resets it to default """

        if not group and name in self.groups:
            raise AttributeError('Failed to clear value (name=%s, group=%s): '
                                 'cannot clear values for a group' %
                                 (name, group))

        item = self.attributes.get((group, name))

        if item['_metadata'].default is None:   # pylint: disable=W0104
            item['value'] = None
        else:
            item['value'] = self._transform(item, item['_metadata'].default)

    def read(self, filename):
        cp = ConfigParser.RawConfigParser() #pylint: disable=C0103
        cp.read(filename)
        for section in cp.sections():
            for key, value in cp.items(section):
                try:
                    self.set_value(key, value, section)
                except AttributeError as err:
                    log.warning('Error detected while reading %s: %s' %
                                (filename, err))
                    continue

runtime = Config()

# Group: default
runtime.add_attribute(StrAttr(
    name='data_root',
    default=INSTALL_PATH,
    environ='ZTPS_DEFAULT_DATAROOT'
))

runtime.add_attribute(StrAttr(
    name='identifier',
    choices=['systemmac', 'serialnumber'],
    default='serialnumber'
))

runtime.add_attribute(StrAttr(
    name='server_url',
    default='http://ztpserver:8080',
    environ='ZTPS_DEFAULT_SERVER'
))

runtime.add_attribute(BoolAttr(
    name='logging',
    default=True,
    environ='ZTPS_DEFAULT_LOGGING'
))

runtime.add_attribute(BoolAttr(
    name='console_logging',
    default=True
))

runtime.add_attribute(StrAttr(
    name='console_logging_format',
    default='%(asctime)s:%(levelname)s:[%(module)s:%(lineno)d] %(message)s',
    environ='ZTPS_CONSOLE_LOGGING_FORMAT'
))

runtime.add_attribute(BoolAttr(
    name='disable_topology_validation',
    default=False
))

# Group: server
runtime.add_attribute(StrAttr(
    name='interface',
    group='server',
    default='0.0.0.0'
))

runtime.add_attribute(IntAttr(
    name='port',
    group='server',
    min_value=1,
    max_value=65534,
    default=8080
))


# Group: bootstrap
runtime.add_attribute(StrAttr(
    name='filename',
    group='bootstrap',
    default='bootstrap',
    environ='ZTPS_BOOTSTRAP_FILENAME'
))

# Group: neighbordb
runtime.add_attribute(StrAttr(
    name='filename',
    group='neighbordb',
    default='neighbordb',
    environ='ZTPS_NEIGHBORDB_FILENAME'
))
